# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime, timezone
from itertools import product
from typing import Callable, Optional


def n_things(count: int, plural_noun: str, singular_noun: Optional[str]=None):
    if count == 1:
        if singular_noun is None:
            return f'1 {plural_noun[0:-1]}'
        else:
            return f'1 {singular_noun}'
    else:
        return f'{count} {plural_noun}'


def parse_datetime(text: str, now_fn: Callable[[], datetime]=datetime.now) -> datetime:
    """
    Parse the given text, accounting for the fact that humans might be typing on the other side.

    :param text: The text to attempt to parse.
    :param now_fn: A function that, when called, provides the current time.
    :return: A ``datetime`` if the string could be parsed.
    """
    now = now_fn()

    dt = _parse_datetime_no_tz(text, now)
    if dt.tzinfo is not None:
        return dt
    elif text.endswith('Z') or text.endswith('z'):
        return dt.replace(tzinfo=timezone.utc)
    else:
        # Strange API, but this is the only safe way to get a local timezone out of the datetime
        # library. We're essentially forcing the returned datetime to carry the same timezone as
        # the current timezone.
        return dt.replace(tzinfo=now.astimezone(tz=now.tzinfo).tzinfo)


def _parse_datetime_no_tz(text: str, now: datetime) -> datetime:
    for combo_format in _ALL_FORMATS:
        try:
            return datetime.strptime(text, combo_format)
        except ValueError:
            pass

    for date_format in _DATE_ONLY_FORMATS:
        try:
            return datetime.strptime(text, date_format)
        except ValueError:
            pass

    for time_format in _TIME_ONLY_FORMATS:
        try:
            return datetime.strptime(text, time_format).replace(year=now.year, month=now.month, day=now.day)
        except ValueError:
            pass

    raise ValueError(f"Couldn't parse {text!r} as a datetime")


_DATE_ONLY_FORMATS = ('%Y-%m-%d', '%Y%m%d', '%y-%m-%d', '%y%m%d')
_TIME_ONLY_FORMATS = tuple(''.join(segments) for segments in product(
    ('%H:%M', '%H:%M:%S', '%H:%M:%S.', '%H:%M:%S.%f', '%I:%M %p', '%I:%M%p', '%I%p'),
    ('', 'Z', 'z', '%z')))

_ALL_FORMATS = [''.join(segments) for segments in product(_DATE_ONLY_FORMATS, (' ', 'T'), _TIME_ONLY_FORMATS)]



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Conversion methods from Ledger API Protobuf-generated types to dazl/Pythonic types.
"""
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, Iterable, List, Optional, Sequence, Union

# noinspection PyPackageRequirements
from google.protobuf.empty_pb2 import Empty

from ... import LOG
from ...model.core import ContractId
from ...model.reading import BaseEvent, TransactionFilter, ContractCreateEvent, \
    TransactionStartEvent, TransactionEndEvent, ContractArchiveEvent, OffsetEvent, \
    ContractExercisedEvent, ActiveContractSetEvent
from ...model.core import Party
from ...model.types import RecordType, Type, VariantType, ContractIdType, ListType, \
    TypeEvaluationContext, type_evaluate_dispatch_default_error, MapType, OptionalType
from ...model.types_store import PackageStore
from ...util.prim_types import to_date, to_datetime, to_hashable, frozendict


@dataclass(frozen=True)
class BaseEventDeserializationContext:
    """
    Attributes required throughout the deserialization of an event stream.
    """
    client: Any
    store: PackageStore
    party: Party
    ledger_id: str

    def offset_event(self, time: datetime, offset: str) -> OffsetEvent:
        return OffsetEvent(self.client, self.party, time, self.ledger_id, self.store, offset)

    def active_contract_set(self, offset: str, workflow_id: str) \
            -> 'ActiveContractSetEventDeserializationContext':
        return ActiveContractSetEventDeserializationContext(
            self.client, self.store, self.party, self.ledger_id,
            offset, workflow_id)

    def transaction(self, time: datetime, offset: str, command_id: str, workflow_id: str) \
            -> 'TransactionEventDeserializationContext':
        return TransactionEventDeserializationContext(
            self.client, self.store, self.party, self.ledger_id,
            time, offset, command_id, workflow_id)


@dataclass(frozen=True)
class ActiveContractSetEventDeserializationContext(BaseEventDeserializationContext):
    """
    Attributes required throughout the deserialization of the Active Contract Set.
    """
    offset: str
    workflow_id: str

    def active_contract_set_event(self, contract_events: 'Sequence[ContractCreateEvent]') \
            -> 'ActiveContractSetEvent':
        return ActiveContractSetEvent(self.client, self.party, None, self.ledger_id, self.store,
                                      self.offset, contract_events)

    def contract_created_event(self, cid, cdata, event_id, witness_parties) -> ContractCreateEvent:
        return ContractCreateEvent(
            client=self.client, party=self.party, time=None, ledger_id=self.ledger_id,
            package_store=self.store, offset=self.offset,
            command_id='', workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties)


@dataclass(frozen=True)
class TransactionEventDeserializationContext(BaseEventDeserializationContext):
    """
    Attributes required throughout the deserialization of an event stream.
    """
    time: datetime
    offset: str
    command_id: str
    workflow_id: str

    def transaction_start_event(self, contract_events) -> TransactionStartEvent:
        return TransactionStartEvent(self.client, self.party, self.time, self.ledger_id, self.store,
                                     self.offset, self.command_id, self.workflow_id, contract_events)

    def transaction_end_event(self, contract_events) -> TransactionEndEvent:
        return TransactionEndEvent(self.client, self.party, self.time, self.ledger_id, self.store,
                                   self.offset, self.command_id, self.workflow_id, contract_events)

    def contract_created_event(self, cid, cdata, event_id, witness_parties) -> ContractCreateEvent:
        return ContractCreateEvent(
            client=self.client, party=self.party, time=self.time, ledger_id=self.ledger_id,
            package_store=self.store, offset=self.offset,
            command_id=self.command_id, workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties)

    def contract_exercised_event(self, cid, cdata, event_id, witness_parties,
                                 contract_creating_event_id: str, choice: str, choice_argument: Any,
                                 acting_parties, consuming, child_event_ids) \
            -> ContractExercisedEvent:
        return ContractExercisedEvent(
            client=self.client, party=self.party, time=self.time, ledger_id=self.ledger_id,
            package_store=self.store, offset=self.offset,
            command_id=self.command_id, workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties,
            contract_creating_event_id=contract_creating_event_id, acting_parties=acting_parties,
            choice=choice, choice_args=choice_argument, consuming=consuming,
            child_event_ids=child_event_ids)

    def contract_archived_event(self, cid, cdata, event_id, witness_parties) \
            -> ContractArchiveEvent:
        return ContractArchiveEvent(
            client=self.client, party=self.party, time=self.time, ledger_id=self.ledger_id,
            package_store=self.store, offset=self.offset,
            command_id=self.command_id, workflow_id=self.workflow_id,
            cid=cid, cdata=cdata, event_id=event_id, witness_parties=witness_parties)


def serialize_transactions_request(transaction_filter: TransactionFilter, party: str) \
        -> 'G.GetTransactionsRequest':
    from . import model as G
    if transaction_filter.current_offset is not None:
        ledger_offset = G.LedgerOffset()
        ledger_offset.absolute = transaction_filter.current_offset
    else:
        ledger_offset = G.LedgerOffset()
        ledger_offset.boundary = 0

    if transaction_filter.destination_offset is not None:
        final_offset = G.LedgerOffset()
        final_offset.absolute = transaction_filter.destination_offset
    else:
        final_offset = G.LedgerOffset()
        final_offset.boundary = 1

    filters_by_party = {party: G.Filters()}

    # add the party groups in the filters by party
    if transaction_filter.party_groups is not None:
        for party_group in transaction_filter.party_groups:
            filters_by_party[party_group] = G.Filters()

    tr_filter = G.TransactionFilter(filters_by_party=filters_by_party)
    return G.GetTransactionsRequest(
        ledger_id=transaction_filter.ledger_id,
        begin=ledger_offset,
        end=final_offset,
        filter=tr_filter)


def serialize_event_id_request(ledger_id: str, event_id: str, requesting_parties: 'Sequence[str]') \
        -> 'G.GetTransactionByEventIdRequest':
    from . import model as G
    return G.GetTransactionByEventIdRequest(ledger_id=ledger_id,
                                            event_id=event_id,
                                            requesting_parties=requesting_parties)


def serialize_acs_request(ledger_id: str, party: str):
    from . import model as G

    filters_by_party = {party: G.Filters()}
    tr_filter = G.TransactionFilter(filters_by_party=filters_by_party)
    return G.GetActiveContractsRequest(
        ledger_id=ledger_id,
        filter=tr_filter)


def to_transaction_events(
        context: 'BaseEventDeserializationContext',
        tx_stream_pb: 'Iterable[G.GetTransactionsResponse]',
        tt_stream_pb: 'Iterable[G.GetTransactionTreesResponse]',
        last_offset_override: 'Optional[str]') -> 'Sequence[BaseEvent]':
    """
    Convert a stream of :class:`GetTransactionsResponse` into a sequence of events.

    :param context:
        Additional data required for deserializing wire information into easily-consumable
        application events.
    :param tx_stream_pb:
        A stream of :class:`GetTransactionsResponse`.
    :param tt_stream_pb:
        A stream of :class:`GetTransactionTreesResponse`. Transactions included in this message
        that are _not_ in ``tx_stream_pb`` will be discarded.
    :param last_offset_override:
        An optional last offset, that, if specified AND different from the last offset of the
        stream, causes a synthetic :class:`OffsetEvent` to be created at the end.
    :return:
    """
    events = []  # type: List[OffsetEvent]

    events_by_offset = dict()  # type: Dict[str, List[OffsetEvent]]
    for item_pb in tx_stream_pb:
        for transaction_pb in item_pb.transactions:
            events_by_offset[transaction_pb.offset] = list(
                to_transaction_chunk(context, transaction_pb))

    for transaction_tree_pb in tt_stream_pb:
        for transaction_pb in transaction_tree_pb.transactions:
            tx_events = events_by_offset.get(transaction_pb.offset)
            if tx_events is not None:
                tx_events[-1:-1] = from_transaction_tree(context, transaction_pb)

    for tx_events in events_by_offset.values():
        events.extend(tx_events)

    if last_offset_override is not None:
        last_time = None if not events else events[-1].time
        last_offset = None if not events else events[-1].offset
        if last_offset != last_offset_override:
            events.append(context.offset_event(last_time, last_offset_override))

    return events


def to_acs_events(
        context: 'BaseEventDeserializationContext',
        acs_stream_pb: 'Iterable[Any]') -> 'Sequence[ActiveContractSetEvent]':
    return [acs_evt
            for acs_response_pb in acs_stream_pb
            for acs_evt in to_acs_event(context, acs_response_pb)]


def from_transaction_tree(context: 'BaseEventDeserializationContext', tt_pb: 'G.TransactionTree') \
        -> 'Sequence[OffsetEvent]':

    t_context = context.transaction(
        time=to_datetime(tt_pb.effective_at),
        offset=tt_pb.offset,
        command_id=tt_pb.command_id,
        workflow_id=tt_pb.workflow_id)

    events = []
    for evt_pb in tt_pb.events_by_id.values():
        evt = to_event(t_context, evt_pb)
        if isinstance(evt, ContractExercisedEvent):
            events.append(evt)

    return events


def to_acs_event(context: BaseEventDeserializationContext, acs_pb: 'G.ActiveContractSetResponse'):
    acs_context = context.active_contract_set(acs_pb.offset, acs_pb.workflow_id)
    contract_events = \
        [evt
         for evt in (to_created_event(acs_context, evt_pb) for evt_pb in acs_pb.active_contracts)
         if evt is not None]
    return [acs_context.active_contract_set_event(contract_events)]


def to_transaction_chunk(context: BaseEventDeserializationContext, tx_pb: 'G.Transaction') \
        -> 'Sequence[OffsetEvent]':
    """
    Return a sequence of events parsed from a Ledger API ``Transaction`` protobuf message.

    :param context:
        Additional information taken from the enclosing message.
    :param tx_pb:
        The Ledger API ``Transaction`` to parse.
    :return:
        A :class:`TransactionStartEvent`, followed by zero or more :class:`ContractCreateEvent`
        and/or :class:`ContractArchiveEvent`, followed by :class:`TransactionEndEvent`.
    """

    t_context = context.transaction(
        time=to_datetime(tx_pb.effective_at),
        offset=tx_pb.offset,
        command_id=tx_pb.command_id,
        workflow_id=tx_pb.workflow_id)

    contract_events = [to_event(t_context, evt_pb) for evt_pb in tx_pb.events]
    contract_events = [evt for evt in contract_events if evt is not None]

    return [t_context.transaction_start_event(contract_events),
            *contract_events,
            t_context.transaction_end_event(contract_events)]


def to_event(
        context: 'Union[TransactionEventDeserializationContext, ActiveContractSetEventDeserializationContext]',
        evt_pb: 'G.Event') \
        -> 'Optional[BaseEvent]':
    try:
        event_type = evt_pb.WhichOneof('event')
    except ValueError:
        try:
            event_type = evt_pb.WhichOneof('kind')
        except ValueError:
            LOG.error('Deserialization error into an event of %r', evt_pb)
            raise

    if 'created' == event_type:
        return to_created_event(context, evt_pb.created)
    elif 'exercised' == event_type:
        return to_exercised_event(context, evt_pb.exercised)
    elif 'archived' == event_type:
        return to_archived_event(context, evt_pb.archived)
    else:
        raise ValueError(f'unknown event type: {event_type}')


def to_created_event(
        context: 'Union[TransactionEventDeserializationContext, ActiveContractSetEventDeserializationContext]',
        cr: 'G.CreatedEvent') \
        -> 'Optional[ContractCreateEvent]':
    search_str = f'{cr.template_id.name}@{cr.template_id.package_id}'

    candidates = context.store.resolve_template_type(search_str)
    if len(candidates) == 0:
        LOG.warning('Could not find metadata for %s!', search_str)
        return None
    elif len(candidates) > 1:
        LOG.error('The template identifier %s is not unique within its metadata!', candidates)
        return None

    ((type_ref, tt),) = candidates.items()

    tt_context = TypeEvaluationContext.from_store(context.store)

    cid = ContractId(cr.contract_id, type_ref)
    cdata = to_record(tt_context, tt, cr.create_arguments)
    event_id = cr.event_id
    witness_parties = tuple(cr.witness_parties)

    return context.contract_created_event(cid, cdata, event_id, witness_parties)


def to_exercised_event(
        context: 'TransactionEventDeserializationContext',
        er: 'G.ExercisedEvent') \
        -> 'Optional[ContractExercisedEvent]':
    search_str = f'{er.template_id.name}@{er.template_id.package_id}'

    candidates = context.store.resolve_template_type(search_str)
    if len(candidates) == 0:
        LOG.warning('Could not find metadata for %s!', search_str)
        return None
    elif len(candidates) > 1:
        LOG.error('The template identifier %s is not unique within its metadata!', candidates)
        return None

    ((type_ref, tt),) = candidates.items()

    tt_context = TypeEvaluationContext.from_store(context.store)
    choice_candidates = context.store.resolve_choice(type_ref, er.choice)
    if len(candidates) == 0:
        LOG.warning('Could not find metadata for %s!', search_str)
        return None
    elif len(candidates) > 1:
        LOG.error('The template identifier %s is not unique within its metadata!', candidates)
        return None

    ((choice_ref, cc),) = choice_candidates.items()

    cid = ContractId(er.contract_id, type_ref)
    event_id = er.event_id
    witness_parties = tuple(er.witness_parties)
    contract_creating_event_id = er.contract_creating_event_id
    choice = er.choice
    choice_args = to_natural_type(tt_context, cc.data_type, er.choice_argument)
    acting_parties = tuple(er.acting_parties)
    consuming = er.consuming
    child_event_ids = er.child_event_ids

    return context.contract_exercised_event(
        cid, None, event_id, witness_parties, contract_creating_event_id,
        choice, choice_args, acting_parties, consuming, child_event_ids)


def to_archived_event(
        context: 'TransactionEventDeserializationContext',
        ar: 'G.ArchivedEvent') \
        -> 'Optional[ContractArchiveEvent]':
    search_str = f'{ar.template_id.name}@{ar.template_id.package_id}'

    candidates = context.store.resolve_template_type(search_str)
    if len(candidates) == 0:
        LOG.warning('Could not find metadata for %s!', search_str)
        return None
    elif len(candidates) > 1:
        LOG.error('The template identifier %s is not unique within its metadata!', candidates)
        return None

    ((type_ref, tt),) = candidates.items()
    event_id = ar.event_id
    witness_parties = tuple(ar.witness_parties)

    cid = ContractId(ar.contract_id, type_ref)
    return context.contract_archived_event(cid, None, event_id, witness_parties)


def to_natural_type(context: TypeEvaluationContext, data_type: Type, obj: 'G.Value') -> Any:
    ctor = obj.WhichOneof('Sum')
    if ctor == 'record':
        return to_record(context, data_type, obj.record)
    elif ctor == 'variant':
        return to_variant(context, data_type, obj.variant)
    elif ctor == 'contract_id':
        return to_contract_id(context, data_type, obj.contract_id)
    elif ctor == 'list':
        return to_list(context, data_type, obj.list)
    elif ctor == 'optional':
        return to_optional(context, data_type, obj.optional)
    elif ctor == 'int64':
        return to_int64(obj.int64)
    elif ctor == 'decimal':
        return to_decimal(obj.decimal)
    elif ctor == 'text':
        return to_text(obj.text)
    elif ctor == 'date':
        return to_date(obj.date)
    elif ctor == 'timestamp':
        return to_datetime(obj.timestamp)
    elif ctor == 'party':
        return to_party(obj.party)
    elif ctor == 'bool':
        return to_bool(obj.bool)
    elif ctor == 'unit':
        return to_unit(obj.unit)
    elif ctor is None:
        return None
    else:
        raise ValueError(f'unhandled value type: {ctor!r}')


def to_record(context: TypeEvaluationContext, tt: Type, record: 'G.Record') -> dict:
    def process(child_context: TypeEvaluationContext, rt: RecordType) -> dict:
        args_list = rt.as_args_list()

        natural = dict()
        for (name, field_type), field in zip(args_list, record.fields):
            natural[name] = to_natural_type(child_context.append_path(name), field_type, field.value)
        return natural

    return type_evaluate_dispatch_default_error(on_record=process)(context, tt)


def to_variant(context: TypeEvaluationContext, tt: Type, variant: 'G.Variant') -> Any:
    """
    Convert an on-the-wire :class:`G.Variant` to a Python type.

    Currently this also handles transparent conversion of :class:`MapType` and :class:`OptionalType`
    to native Python types as well.

    :param context: The :class:`TypeEvaluationContext`
    :param tt: The DAML :class:`Type` to convert to.
    :param variant: The on-the-wire :class:`G.Variant` Protobuf message.
    :return: The native Python representation of this variant.
    """
    def process_variant(child_context: TypeEvaluationContext, vt: VariantType) -> dict:
        ctor = variant.constructor
        field_type = vt.field_type(ctor)
        return {ctor: to_natural_type(child_context.append_path(ctor), field_type, variant.value)}

    def process_map(child_context: TypeEvaluationContext, mt: MapType) -> dict:
        reformatted = {}
        for map_internal_tuple in variant.value.list.elements:
            # TODO: Bake hashability into to_natural_type
            key = to_hashable(to_natural_type(
                child_context, mt.key_type, map_internal_tuple.record.fields[0].value))
            value = to_hashable(to_natural_type(
                child_context, mt.value_type, map_internal_tuple.record.fields[1].value))
            reformatted[key] = value
        return frozendict(reformatted)

    def process_optional(child_context: TypeEvaluationContext, ot: OptionalType) -> Optional[Any]:
        if variant.constructor == 'None':
            return None
        elif variant.constructor == 'Some':
            return to_natural_type(child_context.append_path('?'), ot.type_parameter, variant.value)

    return type_evaluate_dispatch_default_error(
        on_optional=process_optional,
        on_map=process_map,
        on_variant=process_variant)(context, tt)


def to_contract_id(context: TypeEvaluationContext, tt: Type, contract_id: str) \
        -> ContractId:
    def process(_: TypeEvaluationContext, ct: ContractIdType) -> ContractId:
        return ContractId(contract_id, ct.type_parameter)
    return type_evaluate_dispatch_default_error(on_contract_id=process)(context, tt)


def to_list(context: TypeEvaluationContext, tt: Type, list_: 'G.List') -> list:
    def process(child_context: TypeEvaluationContext, lt: ListType) -> list:
        return [to_natural_type(child_context.append_path(f'[{i}]'), lt.type_parameter, element)
                for i, element in enumerate(list_.elements)]
    return type_evaluate_dispatch_default_error(on_list=process)(context, tt)


def to_optional(context: TypeEvaluationContext, tt: Type, optional: 'G.Optional') -> Any:
    def process(child_context: TypeEvaluationContext, ot: OptionalType) -> list:
        return to_natural_type(child_context.append_path(f'?'), ot.type_parameter, optional.value)
    return type_evaluate_dispatch_default_error(on_optional=process)(context, tt)


def to_int64(int64: int) -> int:
    return int64


def to_decimal(decimal: str) -> Decimal:
    return Decimal(decimal)


def to_text(text: str) -> str:
    return text


def to_party(party: str) -> str:
    return party


def to_bool(bool_: bool) -> bool:
    return bool_


def to_unit(_: Empty) -> dict:
    return {}



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ..model.core import DazlError


class SDKComponentUnknownError(DazlError):
    def __init__(self, component_name):
        self.component_name = component_name


class SDKComponentNotFoundError(DazlError):
    def __init__(self, component_name):
        self.component_name = component_name



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import ensure_future, gather, sleep, Future
from typing import Collection, List, Optional, Tuple, TYPE_CHECKING

from .. import LOG
from ..model.reading import sortable_offset_height, max_offset
from ..util.asyncio_util import completed, named_gather

if TYPE_CHECKING:
    from ._party_client_impl import _PartyClientImpl

"""
Functions for ensuring that readers across different parties remain as close in sync to each other
as practical.
"""


async def run_iteration(party_impls: 'Collection[_PartyClientImpl]') \
        -> 'Tuple[str, Collection[Future]]':
    """
    Read the next set of transactions for the set of parties. This coroutine ends when all
    parties are caught up to the same offset.

    :param party_impls:
        A collection of :class:`_PartyClientImpl` will have scheduled invocations.
    :return:
        A tuple of the current ending offset and Futures that represent completions for commands
        that were submitted as a direct result of these events.
    """
    read_coroutines = []  # type: List[Future]

    # have every client read as far ahead as they can
    offsets, event_fut = await read_transactions(party_impls, None, True)
    if not event_fut.done():
        read_coroutines.append(event_fut)

    final_offset = max_offset(offsets)
    if len(offsets) > 1:
        # now have every client catch up to the agreed-upon HEAD
        _, event_fut = await read_transactions(party_impls, final_offset, True)
        if not event_fut.done():
            read_coroutines.append(event_fut)

    await sleep(0)

    return final_offset, [fut for fut in read_coroutines if not fut.done()]


async def read_initial_acs(
        party_impls: 'Collection[_PartyClientImpl]') -> 'Optional[str]':
    """
    Perform the initial synchronization of the Active Contract Set with the server, using the
    Active Contract Set service.

    :param party_impls:
        A collection of _PartyClientImpl's whose offset and ACS state need to be updated.
    :return:
        The ending offset that all parties have reached, or ``None`` if no calls were actually
        made.
    """
    LOG.info('Reading current ledger state from the Active Contract Set service...')

    # Fetch the ACS as every single client.
    if not party_impls:
        return None

    offsets = [offset for offset, _ in
               await gather(*[party_impl.read_acs(None, False) for party_impl in party_impls])]
    offset = max_offset(offsets)

    # Find the most recent offset among all of those clients, and use the transaction stream
    # to catch up all clients to the same point.
    await gather(*[party_impl.read_transactions(offset, False) for party_impl in party_impls])
    return offset


async def read_transaction_event_stream(
        party_impls: 'Collection[_PartyClientImpl]') -> 'Optional[str]':
    LOG.info('Reading current ledger state...')
    if not party_impls:
        return None

    party_impl = next(iter(party_impls))
    offset = await party_impl.read_end()

    await gather(*[party_impl.read_transactions(offset, False) for party_impl in party_impls])
    return offset


async def read_transactions(
        party_impls: 'Collection[_PartyClientImpl]',
        until_offset: 'Optional[str]',
        raise_events: bool) -> 'Tuple[Collection[str], Future]':
    """
    Read transactions from a collection of PartyImpls.

    :param party_impls:
    :param until_offset:
    :param raise_events:
    :return:
        A tuple containing:
         * a set of the offsets returned from all readers, and
         * a Future that is resolved when all events across all readers have resolved either
           successfully or unsuccessfully.
    """
    tuples = await gather(*(pi.read_transactions(until_offset, raise_events) for pi in party_impls))
    offsets = sorted({t[0] for t in tuples}, key=sortable_offset_height)
    futures = [ensure_future(t[1]) for t in tuples]
    futures = [fut for fut in futures if not fut.done()]
    if not futures:
        return offsets, completed(None)
    elif len(futures):
        return offsets, futures[0]
    else:
        return offsets, named_gather(repr(futures), *futures, return_exceptions=True)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/transaction.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import event_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2
from . import trace_context_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/transaction.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B\025TransactionOuterClass'),
  serialized_pb=_b('\n0com/digitalasset/ledger/api/v1/transaction.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a*com/digitalasset/ledger/api/v1/event.proto\x1a\x32\x63om/digitalasset/ledger/api/v1/trace_context.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xa7\x03\n\x0fTransactionTree\x12\x16\n\x0etransaction_id\x18\x01 \x01(\t\x12\x12\n\ncommand_id\x18\x02 \x01(\t\x12\x13\n\x0bworkflow_id\x18\x03 \x01(\t\x12\x30\n\x0c\x65\x66\x66\x65\x63tive_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06offset\x18\x06 \x01(\t\x12U\n\x0c\x65vents_by_id\x18\x07 \x03(\x0b\x32?.com.digitalasset.ledger.api.v1.TransactionTree.EventsByIdEntry\x12\x16\n\x0eroot_event_ids\x18\x08 \x03(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\x1a\\\n\x0f\x45ventsByIdEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).com.digitalasset.ledger.api.v1.TreeEvent:\x02\x38\x01\"\x99\x01\n\tTreeEvent\x12?\n\x07\x63reated\x18\x01 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.CreatedEventH\x00\x12\x43\n\texercised\x18\x02 \x01(\x0b\x32..com.digitalasset.ledger.api.v1.ExercisedEventH\x00\x42\x06\n\x04kind\"\x8d\x02\n\x0bTransaction\x12\x16\n\x0etransaction_id\x18\x01 \x01(\t\x12\x12\n\ncommand_id\x18\x02 \x01(\t\x12\x13\n\x0bworkflow_id\x18\x03 \x01(\t\x12\x30\n\x0c\x65\x66\x66\x65\x63tive_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x35\n\x06\x65vents\x18\x05 \x03(\x0b\x32%.com.digitalasset.ledger.api.v1.Event\x12\x0e\n\x06offset\x18\x06 \x01(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContextB7\n\x1e\x63om.digitalasset.ledger.api.v1B\x15TransactionOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2.DESCRIPTOR,com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])




_TRANSACTIONTREE_EVENTSBYIDENTRY = _descriptor.Descriptor(
  name='EventsByIdEntry',
  full_name='com.digitalasset.ledger.api.v1.TransactionTree.EventsByIdEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='com.digitalasset.ledger.api.v1.TransactionTree.EventsByIdEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.digitalasset.ledger.api.v1.TransactionTree.EventsByIdEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=545,
  serialized_end=637,
)

_TRANSACTIONTREE = _descriptor.Descriptor(
  name='TransactionTree',
  full_name='com.digitalasset.ledger.api.v1.TransactionTree',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='com.digitalasset.ledger.api.v1.TransactionTree.transaction_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='command_id', full_name='com.digitalasset.ledger.api.v1.TransactionTree.command_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='workflow_id', full_name='com.digitalasset.ledger.api.v1.TransactionTree.workflow_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='effective_at', full_name='com.digitalasset.ledger.api.v1.TransactionTree.effective_at', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset', full_name='com.digitalasset.ledger.api.v1.TransactionTree.offset', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='events_by_id', full_name='com.digitalasset.ledger.api.v1.TransactionTree.events_by_id', index=5,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='root_event_ids', full_name='com.digitalasset.ledger.api.v1.TransactionTree.root_event_ids', index=6,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.TransactionTree.trace_context', index=7,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTIONTREE_EVENTSBYIDENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=214,
  serialized_end=637,
)


_TREEEVENT = _descriptor.Descriptor(
  name='TreeEvent',
  full_name='com.digitalasset.ledger.api.v1.TreeEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='created', full_name='com.digitalasset.ledger.api.v1.TreeEvent.created', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exercised', full_name='com.digitalasset.ledger.api.v1.TreeEvent.exercised', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kind', full_name='com.digitalasset.ledger.api.v1.TreeEvent.kind',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=640,
  serialized_end=793,
)


_TRANSACTION = _descriptor.Descriptor(
  name='Transaction',
  full_name='com.digitalasset.ledger.api.v1.Transaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='com.digitalasset.ledger.api.v1.Transaction.transaction_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='command_id', full_name='com.digitalasset.ledger.api.v1.Transaction.command_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='workflow_id', full_name='com.digitalasset.ledger.api.v1.Transaction.workflow_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='effective_at', full_name='com.digitalasset.ledger.api.v1.Transaction.effective_at', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='events', full_name='com.digitalasset.ledger.api.v1.Transaction.events', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset', full_name='com.digitalasset.ledger.api.v1.Transaction.offset', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.Transaction.trace_context', index=6,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=796,
  serialized_end=1065,
)

_TRANSACTIONTREE_EVENTSBYIDENTRY.fields_by_name['value'].message_type = _TREEEVENT
_TRANSACTIONTREE_EVENTSBYIDENTRY.containing_type = _TRANSACTIONTREE
_TRANSACTIONTREE.fields_by_name['effective_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TRANSACTIONTREE.fields_by_name['events_by_id'].message_type = _TRANSACTIONTREE_EVENTSBYIDENTRY
_TRANSACTIONTREE.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_TREEEVENT.fields_by_name['created'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2._CREATEDEVENT
_TREEEVENT.fields_by_name['exercised'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2._EXERCISEDEVENT
_TREEEVENT.oneofs_by_name['kind'].fields.append(
  _TREEEVENT.fields_by_name['created'])
_TREEEVENT.fields_by_name['created'].containing_oneof = _TREEEVENT.oneofs_by_name['kind']
_TREEEVENT.oneofs_by_name['kind'].fields.append(
  _TREEEVENT.fields_by_name['exercised'])
_TREEEVENT.fields_by_name['exercised'].containing_oneof = _TREEEVENT.oneofs_by_name['kind']
_TRANSACTION.fields_by_name['effective_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TRANSACTION.fields_by_name['events'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2._EVENT
_TRANSACTION.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
DESCRIPTOR.message_types_by_name['TransactionTree'] = _TRANSACTIONTREE
DESCRIPTOR.message_types_by_name['TreeEvent'] = _TREEEVENT
DESCRIPTOR.message_types_by_name['Transaction'] = _TRANSACTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TransactionTree = _reflection.GeneratedProtocolMessageType('TransactionTree', (_message.Message,), dict(

  EventsByIdEntry = _reflection.GeneratedProtocolMessageType('EventsByIdEntry', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTIONTREE_EVENTSBYIDENTRY,
    __module__ = 'com.digitalasset.ledger.api.v1.transaction_pb2'
    # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.TransactionTree.EventsByIdEntry)
    ))
  ,
  DESCRIPTOR = _TRANSACTIONTREE,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.TransactionTree)
  ))
_sym_db.RegisterMessage(TransactionTree)
_sym_db.RegisterMessage(TransactionTree.EventsByIdEntry)

TreeEvent = _reflection.GeneratedProtocolMessageType('TreeEvent', (_message.Message,), dict(
  DESCRIPTOR = _TREEEVENT,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.TreeEvent)
  ))
_sym_db.RegisterMessage(TreeEvent)

Transaction = _reflection.GeneratedProtocolMessageType('Transaction', (_message.Message,), dict(
  DESCRIPTOR = _TRANSACTION,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.Transaction)
  ))
_sym_db.RegisterMessage(Transaction)


DESCRIPTOR._options = None
_TRANSACTIONTREE_EVENTSBYIDENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: da/daml_lf_0.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='da/daml_lf_0.proto',
  package='daml_lf_0',
  syntax='proto3',
  serialized_options=_b('\n\030com.digitalasset.daml_lf'),
  serialized_pb=_b('\n\x12\x64\x61/daml_lf_0.proto\x12\tdaml_lf_0\"\t\n\x07PackageB\x1a\n\x18\x63om.digitalasset.daml_lfb\x06proto3')
)




_PACKAGE = _descriptor.Descriptor(
  name='Package',
  full_name='daml_lf_0.Package',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33,
  serialized_end=42,
)

DESCRIPTOR.message_types_by_name['Package'] = _PACKAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Package = _reflection.GeneratedProtocolMessageType('Package', (_message.Message,), dict(
  DESCRIPTOR = _PACKAGE,
  __module__ = 'da.daml_lf_0_pb2'
  # @@protoc_insertion_point(class_scope:daml_lf_0.Package)
  ))
_sym_db.RegisterMessage(Package)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
:mod:`dazl.client` package
==========================

The :mod:`dazl.client` module a high-level view of the Ledger API in a friendly way.

It provides:

  * a callback-based API for interacting with events on the read-side of the Ledger API
  * convenient methods for creating arbitrary commands on the write-side of the Ledger API.

For a higher-level, more declarative API, see :mod:`dazl.query`.
"""

from . import config
from .api import simple_client, Network, PartyClient, AIOPartyClient, SimplePartyClient
from ._base_model import ExitCode, LedgerRun
from .client_participant import ParticipantLedgerClient
from .manager import LedgerClientManager, create_client




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Utilities for dealing with the file system.
"""

import socket
import sys
from io import TextIOBase, SEEK_SET
from types import TracebackType
from typing import TextIO, Optional, Type, Iterator, Iterable, List


def read_file_bytes(file_path: Optional[str]) -> Optional[bytes]:
    """
    Read the contents of a file as ``bytes``. If ``file_path`` is ``None``, then ``None`` is
    returned instead.
    """
    if file_path is not None:
        with open(file_path, 'rb') as f:
            return f.read()
    else:
        return None


def find_free_port() -> int:
    sock = socket.socket()
    try:
        sock.bind(('', 0))
        return sock.getsockname()[1]
    finally:
        sock.close()


def is_port_alive(port: int) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        result = sock.connect_ex(('127.0.0.1', port))
        return result == 0
    finally:
        sock.close()


class LoggingStream(TextIOBase):
    """
    A text "stream" that actually redirects to the specified logger at the specified log level.
    """
    def __init__(self, logger, log_level):
        self.logger = logger
        self.log_level = log_level
        self._buf = None  # type: Optional[str]

    def __enter__(self) -> 'LoggingStream':
        """
        Does nothing; LoggingStreams are not closeable.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Does nothing; LoggingStreams are not closeable.
        """

    @property
    def buffer(self):
        return None

    def closed(self) -> bool:
        return False

    def close(self) -> None:
        """
        Does nothing; LoggingStreams are not closeable.
        """

    def write(self, text: str) -> int:
        # weed out empty or None strings
        if not text:
            return 0

        lines = str.splitlines(text)
        trail = None if text.endswith('\n') else lines.pop()
        if lines:
            # if there is at least one line that we are definitely writing out, then grab the rest
            # of our buffer and prepend it to the first line
            if self._buf is not None:
                prefix = self._buf
                lines[0] = prefix + lines[0]
                self._buf = None

        if trail is not None:
            # if there is trailing text that we are not supposed to write, either append it to our
            # buffer in progress or create a new buffer specifically for that text
            if self._buf is not None:
                self._buf += trail
            else:
                self._buf = trail

        for line in lines:
            # now actually log the lines we were supposed to log
            self.logger.log(self.log_level, line)

        return len(text)

    def writelines(self, lines: Iterable[str]) -> None:
        for line in lines:
            if self._buf is not None:
                line = self._buf + line
                self._buf = None
            self.logger.log(self.log_level, line)


class StdoutStreamWrapper(TextIO):
    """
    Wrapper for ``sys.stdout`` that can cope with it being reassigned at will (like the ``unittest``
    module does). This allows for log output to be more usefully associated with tests.
    """

    def __enter__(self) -> TextIO:
        return sys.stdout.__enter__()

    @property
    def buffer(self):
        return sys.stdout.buffer

    def closed(self):
        return sys.stdout.closed

    def close(self) -> None:
        return sys.stdout.close()

    @property
    def name(self):
        return sys.stdout.name

    @property
    def encoding(self):
        return sys.stdout.encoding

    @property
    def errors(self):
        return sys.stdout.errors

    @property
    def mode(self):
        return sys.stdout.mode

    @property
    def newlines(self):
        return sys.stdout.newlines

    @property
    def line_buffering(self):
        return sys.stdout.line_buffering

    def fileno(self) -> int:
        return sys.stdout.fileno()

    def flush(self) -> None:
        sys.stdout.flush()

    def isatty(self) -> bool:
        return sys.stdout.isatty()

    def read(self, size: int = -1) -> str:
        return sys.stdout.read(size)

    def readable(self) -> bool:
        return sys.stdout.readable()

    def readline(self, limit: int = -1) -> str:
        return sys.stdout.readline(limit)

    def readlines(self, hint: int = -1) -> List[str]:
        return sys.stdout.readlines(hint)

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        return sys.stdout.seek(offset, whence)

    def seekable(self) -> bool:
        return sys.stdout.seekable()

    def tell(self) -> int:
        return sys.stdout.tell()

    def truncate(self, size: Optional[int] = None) -> int:
        return sys.stdout.truncate(size)

    def writable(self) -> bool:
        return sys.stdout.writable()

    def write(self, s: str) -> int:
        return sys.stdout.write(s)

    def writelines(self, lines: Iterable[str]) -> None:
        return sys.stdout.writelines(lines)

    def __next__(self) -> str:
        return next(sys.stdout)

    def __iter__(self) -> Iterator[str]:
        return iter(sys.stdout)

    def __exit__(self, t: Optional[Type[BaseException]], value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        return sys.stdout.__exit__(t, value, traceback)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from contextlib import ExitStack
from pathlib import Path
from typing import Union, List, Sequence

from .. import LOG
from ..model.types_store import PackageStore
from ..util.dar import TemporaryDar, parse_dalf, DarFile
from ..util.path_util import pathify


class LocalDarRepository:
    def __init__(self):
        self._context = ExitStack()
        self._dar_paths = []  # type: List[Path]
        self._files = set()
        self.store = PackageStore.empty()

    def _add_source(self, path: Path) -> None:
        ext = path.suffix.lower()

        if ext == '.daml':
            # dar_create_start_time = time.time()
            dar_paths = self._context.enter_context(TemporaryDar(path))
            self.add_source(*dar_paths)
            # LOG.info('Compiled a dar in % seconds.', time.time() - dar_create_start_time)

        elif ext == '.dalf':
            dalf_package = parse_dalf(path.read_bytes())
            self._dar_paths.append(path)
            self.store.register_all(dalf_package)

        elif ext == '.dar':
            # dar_parse_start_time = time.time()
            dar = self._context.enter_context(DarFile(path))
            dar_package = dar.read_metadata()
            self._dar_paths.append(path)
            self.store.register_all(dar_package)
            # dar_parse_end_time = time.time()
            # LOG.info('Parsed a dar in %s seconds.', dar_parse_end_time - dar_parse_start_time)

        else:
            LOG.error('Unknown extension: %s', ext)
            raise ValueError(f'Unknown extension: {ext}')

    def add_source(self, *files: Union[str, Path]) -> None:
        """
        Add a source file (either a .daml file, .dalf file, or a .dar file).

        Attempts to add the same file more than once will be ignored.

        :param files: Files to add to the archive.
        """
        for file in files:
            path = pathify(file).resolve(strict=True)
            if path not in self._files:
                self._files.add(path)
                self._add_source(path)

        # stores = list()
        # files = list(files)
        # while files:
        #     file = files.pop(0)
        #     ext = pathify(file).suffix.lower()
        #
        #     if ext == '.daml':
        #         # dar_create_start_time = time.time()
        #         dar_paths = self._context.enter_context(TemporaryDar(file))
        #         #LOG.info('Compiled a dar in % seconds.', time.time() - dar_create_start_time)
        #         files.extend(dar_paths)
        #     elif ext == '.dalf':
        #         with open(file, 'rb') as f:
        #             contents = f.read()
        #         stores.append(parse_dalf('pkg0', contents))
        #     elif ext == '.dar':
        #         #dar_parse_start_time = time.time()
        #         dar = self._context.enter_context(DarFile(file))
        #         stores.append(dar.read_metadata())
        #         #dar_parse_end_time = time.time()
        #         #LOG.info('Parsed a dar in %s seconds.', dar_parse_end_time - dar_parse_start_time)
        #     else:
        #         LOG.error('Unknown extension: %s', ext)
        #         raise ValueError(f'Unknown extension: {ext}')
        #
        # reduce(lambda store, other_store: store.register_all(other_store), stores, self.store)

    def get_daml_archives(self) -> Sequence[Path]:
        return self._dar_paths

    def __enter__(self):
        """
        Does nothing.
        """
        self._context.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Delete all managed resources in the reverse order that they were created.
        """
        self._context.__exit__(exc_type, exc_val, exc_tb)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import abc
import argparse

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import LedgerClientManager


class Plugin(abc.ABC):
    """
    Definition of a plugin on the DAZL client library.
    """

    def install(self, manager: 'LedgerClientManager') -> None:
        """
        Called in order to have this plugin register its event handlers with the manager.

        :param manager: The ledger client manager to register into.
        """

    def provide_args(self, arg_parser: 'argparse._ActionsContainer') -> None:
        """
        If implemented, load arguments specific to this plugin into the argument parser.

        :param arg_parser: The argument parser to augment.
        """



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import datetime
from asyncio import AbstractEventLoop
from concurrent.futures import Executor
from threading import Event, Thread, RLock
from typing import Awaitable, Dict, Iterable, Optional, Union

from .. import LOG
from ._base import LedgerNetwork, LedgerClient, LedgerConnectionOptions, _LedgerConnectionContext
from .v1.grpc import GRPCv1Connection
from .oauth import oauth_flow
from ..model.core import Party
from ..model.ledger import LedgerMetadata
from ..model.network import HTTPConnectionSettings


class AutodetectLedgerNetwork(LedgerNetwork):
    """
    Auto-detecting pool implementation that automatically routes to the correct pool
    implementation based on the scheme specified in the URL.
    """

    def __init__(self,
                 options: LedgerConnectionOptions,
                 loop: Optional[AbstractEventLoop] = None,
                 executor: Optional[Executor] = None):
        self._options = options
        self._context = _LedgerConnectionContext(options, loop, executor)

        self._closed = False
        self._first_connection_evt = Event()
        self._first_connection = None  # type: Optional[AutodetectConnection]
        self._connections = dict()  # type: Dict[HTTPConnectionSettings, AutodetectConnection]
        self._lock = RLock()
        self._ledger_future = self._context.loop.create_future()
        self._admin_url = None  # type: Optional[str]
        self._main_thread = Thread(target=self._main, daemon=True)
        self._main_thread.start()

    def ledger(self) -> Awaitable[LedgerMetadata]:
        return self._ledger_future

    async def connect(self,
                      party: 'Union[str, Party]',
                      settings: 'HTTPConnectionSettings',
                      context_path: 'Optional[str]' = None,
                      admin_url: 'Optional[str]' = None) -> LedgerClient:
        LOG.info('Establishing a connection to %s on party %s...', settings, party)
        if admin_url is not None:
            self._admin_url = admin_url
        if settings.oauth:
            new_oauth_settings = await oauth_flow(settings.oauth)
            settings = settings._replace(oauth=new_oauth_settings)
        conn = self._get_connection(settings, context_path)
        ledger = await self.ledger()
        if ledger.protocol_version == 'v1':
            from .v1.grpc import GRPCv1LedgerClient
            return GRPCv1LedgerClient(conn, ledger, party)
        elif ledger.protocol_version == 'v0':
            raise RuntimeError(f'Unsupported protocol version: {ledger.protocol_version}')
        else:
            raise RuntimeError(f'Unknown protocol version: {ledger.protocol_version}')

    async def upload_archive(self, contents: bytes) -> None:
        if self._admin_url is None:
            raise RuntimeError('upload_archive unsupported unless admin_url is specified')

        from .admin.http_admin import admin_upload_package
        return await self._context.run_in_background(
            lambda: admin_upload_package(self._admin_url, contents))

    async def set_time(self, new_time: datetime):
        ledger = await self.ledger()
        if ledger.protocol_version == 'v1':
            from .v1.grpc import grpc_set_time
            return await self._context.run_in_background(
                lambda: grpc_set_time(self._first_connection, ledger.ledger_id, new_time))
        elif ledger.protocol_version == 'v0':
            raise RuntimeError(f'Unsupported protocol version: {ledger.protocol_version}')
        else:
            raise RuntimeError(f'Unknown protocol version: {ledger.protocol_version}')

    async def close(self) -> None:
        with self._lock:
            connections = list(self._connections.values())
            self._connections.clear()

        for connection in connections:
            try:
                connection.close()
            except:
                LOG.exception('Had trouble closing a connection.')

    def _close_all(self) -> None:
        LOG.debug('Closing all connections...')
        try:
            with self._lock:
                connections = list(self._connections.values())
                self._connections.clear()

            for connection in connections:
                try:
                    connection.close()
                except:
                    LOG.exception('Had trouble closing a connection.')
        finally:
            LOG.debug('Marked the connection pool as closed.')
            self._closed = True

    @property
    def closed(self):
        return self._closed

    def _get_connection(self, settings: HTTPConnectionSettings, context_path: Optional[str]) \
            -> 'AutodetectConnection':

        if not self._lock.acquire(timeout=5):
            raise TimeoutError('Could not acquire an internal lock quickly enough.')
        try:
            if self._closed:
                raise IOError('LedgerNetwork.close() has already been called')

            if not self._connections:
                LOG.debug('Initializing the first connection to %s...', settings)
                self._first_connection = AutodetectConnection(self._context, settings, context_path)
                self._connections[settings] = self._first_connection
                self._first_connection_evt.set()
                return self._first_connection
            else:
                LOG.debug('Initializing a connection to %s...', settings)
                stub = self._connections.get(settings)
                if stub is None:
                    stub = AutodetectConnection(self._context, settings, context_path)
                    self._connections[settings] = stub
                return stub
        except:
            LOG.exception('An error occurred trying to create a connection.')
            raise
        finally:
            self._lock.release()

    def _main(self) -> None:
        timeout = self._options.connect_timeout
        try:
            LOG.debug('Waiting for the first connection to be established (timeout: %s)...', timeout)
            if not self._first_connection_evt.wait(timeout=timeout.total_seconds()):
                LOG.error('Waited %s for the first connection but it never came. Aborting...', timeout)
                raise Exception('first connection timeout')

            LOG.debug('Starting ledger detection.')
            conn = self._first_connection
            if conn is not None:
                for i, metadata in enumerate(_monitor_ledger_network(conn)):
                    if i == 0:
                        self._context.run_on_loop(lambda: self._ledger_future.set_result(metadata))
                    elif metadata is not None:
                        LOG.warning('The network monitor thread emitted multiple metadata!')
        except Exception as ex:
            LOG.exception('The main monitoring thread died.')

            def _maybe_apply_error():
                if not self._ledger_future.done():
                    LOG.error('The above error was propagated as an initialization error.')
                    self._ledger_future.set_exception(ex)

            # if we haven't even been able to initialize, propagate this error to the init future
            self._context.run_on_loop(_maybe_apply_error)

        finally:
            self._close_all()


class AutodetectConnection(GRPCv1Connection):
    pass


def _monitor_ledger_network(connection: AutodetectConnection) -> Iterable[LedgerMetadata]:
    """
    Monitor the very first connection established and provide general information about
    the ledger to other parties.
    """
    from .v1.grpc import grpc_detect_ledger_id

    ledger_id = grpc_detect_ledger_id(connection)
    if ledger_id is not None:
        LOG.info('Ledger ID: %s', ledger_id)
        from .v1.grpc import grpc_main_thread
        return grpc_main_thread(connection=connection, ledger_id=ledger_id)
    elif connection.settings.scheme in ('grpc', 'grpcs'):
        raise Exception('The protocol was specified as gRPC, but the backend does not support it.')
    else:
        LOG.error("Couldn't connect over gRPC.")
        raise Exception('Connection to the v0 REST API is no longer supported.')



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains supporting infrastructure for built-in method definitions for interpreting
DAML-LF files.
"""


from typing import Any, Optional, Sequence, Union

from .daml_lf_1 import Expr, Type, BuiltinFunction
from ..model.types import TypeReference, ValueReference


class _BuiltinMeta(type):
    def __new__(mcs, name, bases, dct):
        if name != 'Builtin':
            proposed_name = dct.get('name', None)
            proposed_builtin = dct.get('builtin', None)
            if proposed_name is None:
                if proposed_builtin is None:
                    raise ValueError('Neither `name` nor `builtin` are specified')
            elif proposed_builtin is not None:
                raise ValueError('Only `name` or `builtin` can be specified, not both')
            dct['name'] = proposed_name
            dct['builtin'] = proposed_builtin
            args = dct.get('args')
            if args is None:
                raise ValueError(f'`args` must be specified for class {proposed_name}')
            if isinstance(args, Type):
                dct['args'] = (args,)
            elif not isinstance(args, tuple):
                dct['args'] = tuple(args)
        return super().__new__(mcs, name, bases, dct)


class Builtin(metaclass=_BuiltinMeta):
    name: Optional[str]
    builtin: Optional[BuiltinFunction]
    arg_types: Sequence[Type]

    def evaluate(self, type_args: 'Sequence[Type]', args: 'Sequence[Any]') -> 'Any':
        raise NotImplementedError()

    def simplify(self, type_args: 'Sequence[Type]', args: 'Sequence[Expr]') -> 'Expr':
        raise NotImplementedError()


class BuiltinTable:
    def __init__(self):
        self.by_name = dict()
        self.by_builtin = dict()

    def add(self, builtin: Builtin):
        if builtin.builtin is not None:
            self.by_builtin[builtin.builtin] = builtin
        elif builtin.name is not None:
            self.by_name[builtin.name] = builtin
        else:
            raise ValueError(f'A builtin could not be registered! {builtin!r}')

    def resolve(self, ref: 'Union[str, ValueReference, TypeReference, BuiltinFunction]') -> \
            'Optional[Builtin]':
        """
        Return a :class:`Builtin` implementation for the name or reference if one is defined.
        """
        if isinstance(ref, BuiltinFunction):
            # All BuiltinFunctions MUST be defined
            return self.by_builtin[ref]
        elif isinstance(ref, (ValueReference, TypeReference)):
            return self.by_name.get(ref.full_name_unambiguous)
        elif isinstance(ref, str):
            return self.by_name.get(ref)
        else:
            raise TypeError(f'unexpected key type: {ref!r}')

    def __contains__(self, ref):
        if isinstance(ref, BuiltinFunction):
            return True



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from argparse import ArgumentParser
from pathlib import Path

from ._base import CliCommand


class PackageDarCommand(CliCommand):
    name = 'package'
    hidden = True

    def parser(self):
        arg_parser = ArgumentParser('dazl package')
        arg_parser.add_argument(
            'file', metavar='DAML_FILE', help='A .daml file to compile into a package')
        arg_parser.add_argument(
            '--out', '-o', metavar='DAR_FILE', default=None)
        arg_parser.add_argument(
            '--log-level')
        return arg_parser

    def execute(self, args):
        from ..util.dar import build_dar, DamlcPackageError

        root_daml_file = Path(args.file)
        output_dar_file = args.out or root_daml_file.with_suffix('.dar')
        try:
            build_dar(str(root_daml_file), str(output_dar_file))
            return 0
        except DamlcPackageError as ex:
            return ex.exit_code



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
:mod:`dazl.model` package
=========================

The :mod:`dazl.model` module is the low-level domain model for the Ledger. Most of the time, these
classes will not be instantiated directly, but will frequently be passed to your code to work with.

The most important classes:

:mod:`dazl.model.core`:
    :class:`~dazl.model.core.ContractId`

:mod:`dazl.model.writing`:
    Subclasses of :class:`~dazl.model.writing.Command` for the Write API:
    :class:`~dazl.model.writing.CreateCommand` and
    :class:`~dazl.model.writing.ExerciseCommand`

:mod:`dazl.model.types`:
    Type description classes:
    Subclasses of :class:`~dazl.model.types.Type`:
    :class:`~dazl.model.types.ScalarType`,
    :class:`~dazl.model.types.ListType`,
    :class:`~dazl.model.types.RecordType`,
    :class:`~dazl.model.types.VariantType`

.. automodule:: dazl.model.core
.. automodule:: dazl.model.writing
.. automodule:: dazl.model.types

"""

from . import core, ledger, network, reading, types, writing



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, Sequence

from ..damlast.daml_lf_1 import Expr, FieldWithType
from .types import ModuleRef


class DamlDeclaration:
    """

    """
    module: ModuleRef
    name: Sequence[str]


class DamlDataType(DamlDeclaration):
    """

    """
    module: ModuleRef
    name: Sequence[str]
    fields: 'Sequence[FieldWithType]'
    eq_instance: 'Optional[Expr]'
    ord_instance: 'Optional[Expr]'
    show_instance: 'Optional[Expr]'


class DamlTemplate(DamlDataType):
    signatories: 'Expr'
    observers: 'Expr'
    agreement: 'Expr'
    precondition: 'Expr'
    choices: 'Sequence[DamlChoice]'


class DamlChoice:
    name: str



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from asyncio import gather, ensure_future, sleep, Future
from collections import defaultdict
from datetime import timedelta, datetime
from threading import RLock, Thread
from typing import Any, Callable, Dict, Optional, TypeVar, Collection, Awaitable, Set, Tuple, \
    Union, overload

from dataclasses import asdict

from .. import LOG
from ._party_client_impl import _PartyClientImpl
from ._run_level import RunState
from ..client.config import NetworkConfig
from ..metrics import MetricEvents
from ..model.core import Party, RunLevel
from ..model.ledger import LedgerMetadata
from ..model.reading import InitEvent, ReadyEvent, BaseEvent, EventKey
from ..protocols import LedgerNetwork
from ..protocols.autodetect import AutodetectLedgerNetwork
from ..util.asyncio_util import execute_in_loop, completed, Invoker
from ..util.dar import get_archives
from ..util.events import CallbackManager
from ..util.prim_types import to_timedelta, TimeDeltaConvertible


T = TypeVar('T')


class _NetworkImpl:

    __slots__ = ('_lock', '_main_thread', 'invoker', '_party_clients', '_global_impls',
                 '_party_impls', '_run_state', '_callbacks', '_config', '_pool', '_cached_metadata',
                 '_metrics')

    def __init__(self, metrics: 'Optional[MetricEvents]' = None):
        self.invoker = Invoker()
        self._lock = RLock()
        self._main_thread = None  # type: Optional[Thread]
        self._run_state = None  # type: Optional[RunState]
        self._pool = None  # type: Optional[LedgerNetwork]
        if metrics is None:
            # create a default set of metrics
            try:
                # noinspection PyUnresolvedReferences
                from ..metrics.prometheus import PrometheusMetricEvents
                self._metrics = PrometheusMetricEvents.default()
            except ImportError:
                self._metrics = MetricEvents()

        # Separately, keep a reference to metadata because it's accessed quite frequently and it is
        # useful for this to be available even when the pool is closed
        self._cached_metadata = None  # type: Optional[LedgerMetadata]

        self._party_clients = defaultdict(dict)  # type: Dict[type, Dict[Party, Any]]
        self._global_impls = dict()  # type: Dict[type, Any]
        self._party_impls = dict()  # type: Dict[Party, _PartyClientImpl]

        self._config = dict()
        self._callbacks = CallbackManager()

    @overload
    def run_in_loop_threadsafe(self,
                               cb: Callable[[], Union[None, Awaitable[None]]],
                               timeout: float = 30) -> None: ...

    @overload
    def run_in_loop_threadsafe(self,
                               cb: Callable[[], Union[Awaitable[T], T]],
                               timeout: float = 30) -> T: ...

    def run_in_loop_threadsafe(self, cb, timeout=30):
        """
        Schedule a callback to be run on the event loop. This can either be a normal function or a
        coroutine.

        :param cb: The callback to invoke.
        :param timeout: The timeout, in seconds, to abort.
        :return: The returned value from the function or coroutine.
        """
        return self.invoker.run_in_loop(cb, timeout=timeout)

    def set_config(self, *configs: 'NetworkConfig', **kwargs):
        for config in configs:
            self._config.update({k: v for k, v in asdict(config).items() if v is not None})
        self._config.update({k: v for k, v in kwargs.items() if v is not None})

    async def aio_run(self, *coroutines, run_state: Optional[RunState] = None) -> None:
        """
        Coroutine where all network activity is scheduled from.

        :param coroutines:
            Optional additional coroutines to run. ``aio_run`` is only considered complete once all
            of the additional coroutines are also finished.
        :param run_state:
            :class:`RunState` that specifies when the loop is to be terminated. If ``None`` is
            supplied, a default :class:`RunState` that runs until explicitly stopped is assigned.
        """
        from ..metrics.instrumenters import AioLoopPerfMonitor
        from ..protocols import LedgerConnectionOptions

        if run_state is None:
            run_state = RunState(RunLevel.RUN_FOREVER)

        with self._lock:
            self._run_state = run_state
            self.invoker.set_context_as_current()
            base_config = dict(self._config)

        site = None
        with AioLoopPerfMonitor(self._metrics.loop_responsiveness):
            server_port = base_config.get('server_port')
            if server_port is not None:
                LOG.debug('Opening port %s for metrics...', server_port)
                from aiohttp import web
                from ..server import get_app
                app = get_app()
                runner = web.AppRunner(app)
                await runner.setup()
                site = web.TCPSite(runner, '0.0.0.0', server_port)
                ensure_future(site.start())
                LOG.info('Listening on port %s for metrics.', server_port)

            options = LedgerConnectionOptions(connect_timeout=timedelta(seconds=30))

            self._pool = pool = AutodetectLedgerNetwork(
                options=options, loop=self.invoker.get_loop(), executor=self.invoker.get_executor())

            try:
                runner = _NetworkRunner(pool, run_state, self, coroutines)
                await runner.run(base_config)
            finally:
                await self._pool.close()
                if site is not None:
                    await site.stop()
                self._pool = None

    def start(self, run_state: RunState, daemon: bool) -> None:
        """
        Create a background thread, start an event loop, and run the application.
        """
        from asyncio import new_event_loop, set_event_loop

        def background_main():
            LOG.info('Starting an event loop on a background thread...')

            try:
                loop = new_event_loop()
                set_event_loop(loop)

                loop.run_until_complete(self.aio_run(run_state=run_state))
            except:
                LOG.exception('The main event loop died!')

        with self._lock:
            if self._main_thread is not None:
                raise RuntimeError('start() called more than once')
            self._main_thread = Thread(
                target=background_main, daemon=daemon, name=f'dazl:main-{id(self):016x}')

        self._main_thread.start()

    def shutdown(self) -> None:
        """
        Indicate that the client and its connected parties should be stopped as soon as is cleanly
        possible.

        This method can safely be called from any thread, but it must be called after start().
        """
        with self._lock:
            LOG.info('Shutting down...')
            loop = self.invoker.get_loop()
            run_state = self._run_state
            if loop is None:
                raise RuntimeError('shutdown() called on a non-stopped Network')
        loop.call_soon_threadsafe(run_state.handle_sigint)

    def abort(self) -> None:
        """
        Indicate that the client and its connected parties should be stopped as immediately.

        This method can safely be called from any thread, but it must be called after start().
        """
        with self._lock:
            LOG.info('Aborting...')
            loop = self.invoker.get_loop()
            run_state = self._run_state
            if loop is None:
                raise RuntimeError('abort() called on a non-stopped Network')
        loop.call_soon_threadsafe(run_state.handle_sigquit)

    def join(self, timeout=None):
        """

        :param timeout:
        :return:
        """
        with self._lock:
            thread = self._main_thread

        if thread is not None:
            thread.join(timeout=timeout)

    def global_impl(self, ctor: 'Callable[[_NetworkImpl], T]') -> T:
        with self._lock:
            inst = self._global_impls.get(ctor)
            if inst is None:
                inst = ctor(self)
                self._global_impls[ctor] = inst
            return inst

    def party_impl(self, party: Party, ctor: 'Callable[[_PartyClientImpl], T]') -> T:
        with self._lock:
            impl = self._party_impls.get(party)
            if impl is None:
                impl = _PartyClientImpl(self._metrics, self.invoker, party)
                self._party_impls[party] = impl

            mm = self._party_clients[ctor]
            inst = mm.get(party)
            if inst is None:
                inst = ctor(impl)
                mm[party] = inst
            return inst

    def parties(self) -> Collection[Party]:
        """
        Return a snapshot of the set of parties that exist right now.
        """
        with self._lock:
            return list(self._party_impls)

    def party_impls(self) -> 'Collection[_PartyClientImpl]':
        with self._lock:
            return list(self._party_impls.values())

    def simple_metadata(self, timeout: Optional[float] = 30.0) -> LedgerMetadata:
        """
        Retrieve metadata about the ledger.

        If you are using the ``asyncio``-flavored version of the API, you _must_ use
        :meth:`aio_metadata` instead.
        """
        if self._cached_metadata is not None:
            return self._cached_metadata

        with self._lock:
            return execute_in_loop(self.invoker.get_loop(), self._pool.ledger, timeout=timeout)

    async def aio_metadata(self) -> LedgerMetadata:
        """
        Coroutine version of :meth:`metadata`.
        """
        return await self._pool.ledger()

    async def ensure_packages(self, contents: bytes, timeout: TimeDeltaConvertible) -> None:
        """
        Ensure packages specified by the given byte array are loaded on the remote server.

        If an admin URL is specified, dazl will attempt to upload packages to that URL, and wait
        for the package to be reported by the package service.

        :param contents: Contents to attempt to upload.
        :param timeout: Length of time before giving up.
        """
        archives = get_archives(contents)
        admin_url = self._config.get('admin_url')
        if admin_url is not None:
            for archive_bytes in archives.values():
                self._pool.run_in_executor

        else:
            await self.ensure_package_ids(archives.keys(), timeout)

    async def ensure_package_ids(
            self, package_ids: 'Collection[str]', timeout: 'TimeDeltaConvertible'):
        from asyncio import wait_for, TimeoutError
        timeout = to_timedelta(timeout)
        expire_time = datetime.utcnow() + timeout
        metadata = await wait_for(self.aio_metadata(), timeout.total_seconds())
        package_id_set = set(package_ids)
        while datetime.utcnow() < expire_time:
            if package_id_set.issubset(set(metadata.store.package_ids())):
                return
        raise TimeoutError()

    async def get_time(self) -> datetime:
        metadata = await self._pool.ledger()
        await self._pool.sync_time()
        return metadata.time_model.get_time()

    def set_time(self, new_datetime: datetime) -> Awaitable[None]:
        return self._pool.set_time(new_datetime)

    # region Event Handler Management

    # noinspection PyShadowingBuiltins

    def add_event_handler(self, key, handler: Callable[[BaseEvent], None], context):
        """
        Add an event handler to a specific event. Unlike event listeners on party clients, these
        event handlers are not allowed to return anything in response to handling an event.
        """
        self._callbacks.add_listener(key, handler, None, context)

    def emit_event(self, data: BaseEvent):
        for key in EventKey.from_event(data):
            self._callbacks.for_listeners(key)(data)

    # endregion


class _NetworkRunner:
    def __init__(self, pool: LedgerNetwork, run_state: RunState, network_impl: '_NetworkImpl', user_coroutines):
        self.pool = pool
        self.run_state = run_state
        self.initialized_parties = set()  # type: Set[Party]
        self._network_impl = network_impl
        self._read_completions = []
        self._user_coroutines = [ensure_future(coro) for coro in user_coroutines]
        self._writers = []

    async def run(self, base_config: dict):
        LOG.debug('Running a Network with base config: %r', base_config)
        prev_offset = None
        keep_running = True
        while keep_running:
            party_impls = {pi.party: pi for pi in self._network_impl.party_impls()}

            # make sure that all parties are initialized; we do this on EVERY tick because new
            # parties might come later and they'll need to be initialized too
            uninitialized_parties = set(party_impls) - self.initialized_parties
            # TODO: Also handle removed parties at some point
            if uninitialized_parties:
                await self._init([party_impls[party] for party in uninitialized_parties], prev_offset, base_config)
            self.initialized_parties |= uninitialized_parties

            current_offset, keep_running = await self.beat(party_impls.values())
            if keep_running:
                if prev_offset == current_offset:
                    await sleep(0.5)
                else:
                    prev_offset = current_offset

        # prohibit any more command submissions
        LOG.debug('network_run: stopping its writers gracefully...')
        # TODO: This needs to be coordinated better among the list of writers we ACTUALLY started
        for pi in self._network_impl.party_impls():
            pi.stop_writer()

        # wait for writers to finish their writing
        await gather(*self._writers, return_exceptions=True)
        LOG.info('network_run: finished.')
        self._writers = []

    async def beat(self, party_impls: 'Collection[_PartyClientImpl]') -> Tuple[str, bool]:
        """
        Called periodically to schedule reads and writes.
        """
        from ._reader_sync import run_iteration
        offset, completions = await run_iteration(party_impls)

        self._read_completions = [fut for fut in (*self._read_completions, *completions)
                                  if not fut.done()]
        self._user_coroutines = [fut for fut in self._user_coroutines if not fut.done()]

        # If there are no more commands in flight, there is no more activity
        if self.run_state.level >= RunLevel.TERMINATE_GRACEFULLY:
            LOG.info('network_run terminating on user request...')
            return offset, False
        elif self.run_state.level >= RunLevel.RUN_UNTIL_IDLE:
            if not self._read_completions and not self._user_coroutines:
                if all(pi.writer_idle() for pi in party_impls):
                    LOG.info('network_run: terminating because all writers are idle and all '
                             'clients have caught up to offset %s', offset)
                    return offset, False

        return offset, True

    async def _init(
            self, party_impls: 'Collection[_PartyClientImpl]',
            first_offset: 'Optional[str]', base_config: dict) -> Tuple[str, Future]:
        # region log
        if LOG.isEnabledFor(logging.DEBUG):
            parties = [pi.party for pi in party_impls]
            if first_offset is None:
                LOG.debug('Starting first-time initialization for parties: %s', parties)
            else:
                LOG.debug('Starting initialization for parties: %s, advancing to offset: %s', parties, first_offset)
        # endregion
        futs = []

        # establish ledger connections
        if party_impls:
            for party_impl in party_impls:
                party_impl.set_config()
            await gather(*(party_impl.connect_in(self.pool, base_config) for party_impl in party_impls))
        elif first_offset is None:
            # establish a single null connection that will be used only to ensure that metadata can be
            # fetched
            party_impl = _PartyClientImpl(self._network_impl._metrics, self._network_impl.invoker, Party('\x00'))
            await party_impl.connect_in(self.pool, base_config)
        else:
            # trivially, there is nothing to do for init because there are no parties to initialize
            # and this isn't our first rodeo
            return first_offset, completed(None)

        # get metadata about the ledger
        metadata = await self.pool.ledger()
        self._network_impl._cached_metadata = metadata

        # Raise the 'init' event.
        current_time = metadata.time_model.get_time()
        if first_offset is None:
            evt = InitEvent(None, None, current_time, metadata.ledger_id, metadata.store)
            self._network_impl.emit_event(evt)
        for party_impl in party_impls:
            futs.append(party_impl.initialize(current_time, metadata))

        from ._reader_sync import read_transaction_event_stream
        offset = await read_transaction_event_stream(party_impls)

        LOG.debug('Preparing to raise the "ready" event...')
        # Raise the 'ready' event.
        current_time = metadata.time_model.get_time()
        if first_offset is None:
            evt = ReadyEvent(None, None, current_time, metadata.ledger_id, metadata.store, offset)
            self._network_impl.emit_event(evt)
        for party_impl in party_impls:
            evt = ReadyEvent(party_impl, party_impl.party, current_time, metadata.ledger_id,
                             metadata.store, offset)
            futs.append(party_impl.emit_event(evt))
        for party_impl in party_impls:
            party_impl.ready().set_result(None)

        self._writers.extend(ensure_future(party_impl.main_writer()) for party_impl in party_impls)

        # Return the metadata offset for later
        LOG.debug('Caught up to %s.', offset)
        return offset, gather(*futs, return_exceptions=True)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from pygments.lexers.haskell import HaskellLexer
from pygments.lexer import inherit
from pygments.token import Keyword


class DAMLLexer(HaskellLexer):

    name = 'DAML'
    aliases = ['daml']
    filenames = ['*.daml']

    daml_reserved = ('template', 'with', 'controller', 'can', 'ensure', 'daml', 'observer', 'signatory', 'agreement', 'controller', 'nonconsuming', 'return', 'this')

    tokens = {
        'root': [
            (r'\b(%s)(?!\')\b' % '|'.join(daml_reserved), Keyword.Reserved),
            (r'\b(True|False)\b', Keyword.Constant),
            inherit
        ]
    }






# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import party_management_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2


class PartyManagementServiceStub(object):
  """Status: experimental interface, will change before it is deemed production
  ready

  Inspect the party management state of a ledger participant and modify the
  parts that are modifiable. We use 'backing participant' to refer to this
  specific participant in the methods of this API.
  When the participant is run in mode requiring authentication, all the calls 
  in this interface will respond with UNAUTHENTICATED, if the caller fails
  to provide a valid access token, and will respond with PERMISSION_DENIED, if
  the claims in the token are insufficient to perform a given operation.
  Subsequently, only specific errors of individual calls not related to 
  authorization will be described.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetParticipantId = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.admin.PartyManagementService/GetParticipantId',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdResponse.FromString,
        )
    self.ListKnownParties = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.admin.PartyManagementService/ListKnownParties',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesResponse.FromString,
        )
    self.AllocateParty = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.admin.PartyManagementService/AllocateParty',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyResponse.FromString,
        )


class PartyManagementServiceServicer(object):
  """Status: experimental interface, will change before it is deemed production
  ready

  Inspect the party management state of a ledger participant and modify the
  parts that are modifiable. We use 'backing participant' to refer to this
  specific participant in the methods of this API.
  When the participant is run in mode requiring authentication, all the calls 
  in this interface will respond with UNAUTHENTICATED, if the caller fails
  to provide a valid access token, and will respond with PERMISSION_DENIED, if
  the claims in the token are insufficient to perform a given operation.
  Subsequently, only specific errors of individual calls not related to 
  authorization will be described.
  """

  def GetParticipantId(self, request, context):
    """Return the identifier of the backing participant.
    All horizontally scaled replicas should return the same id.
    This method is expected to succeed provided the backing participant is 
    healthy, otherwise it responds with INTERNAL grpc error.
    daml-on-sql: returns an identifier supplied on command line at launch time
    daml-on-kv-ledger: as above
    canton: returns globally unique identifier of the backing participant
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListKnownParties(self, request, context):
    """List the parties known by the backing participant.
    The list returned contains parties whose ledger access is facilitated by
    backing participant and the ones maintained elsewhere.
    This request will always succeed.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AllocateParty(self, request, context):
    """Adds a new party to the set managed by the backing participant.
    Caller specifies a party identifier suggestion, the actual identifier
    allocated might be different and is implementation specific.
    This call will either succeed or respond with UNIMPLEMENTED if synchronous
    party allocation is not supported by the backing participant.
    daml-on-sql: suggestion's uniqueness is checked and call rejected if the
    identifier is already present
    daml-on-kv-ledger: suggestion's uniqueness is checked bby the validators in
    the consensus layer and call rejected if the identifier is already present.
    canton: completely different globally unique identifier is allocated.
    Behind the scenes calls to an internal protocol are made. As that protocol
    is richer than the the surface protocol, the arguments take implicit values
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PartyManagementServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetParticipantId': grpc.unary_unary_rpc_method_handler(
          servicer.GetParticipantId,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.GetParticipantIdResponse.SerializeToString,
      ),
      'ListKnownParties': grpc.unary_unary_rpc_method_handler(
          servicer.ListKnownParties,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.ListKnownPartiesResponse.SerializeToString,
      ),
      'AllocateParty': grpc.unary_unary_rpc_method_handler(
          servicer.AllocateParty,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_party__management__service__pb2.AllocatePartyResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.admin.PartyManagementService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import sys
import time
from collections import OrderedDict, defaultdict
from typing import Any, Collection, Mapping, Set, Union

from toposort import toposort_flatten

from ... import LOG
from ...damlast.daml_lf_1 import DefDataType, Archive
from ...damlast.types import get_old_type
from ...model.types import TypeReference, RecordType, VariantType, SCALAR_TYPE_UNIT, \
    NamedArgumentList, TypeVariable, ModuleRef, TemplateChoice, Template, TypeAdjective, \
    ScalarType, ValueReference
from ...model.types_store import PackageStore, PackageStoreBuilder


def parse_archive_payload(raw_bytes: bytes) -> 'G.ArchivePayload':
    """
    Convert ``bytes`` into a :class:`G.ArchivePayload`.

    Note that this function will temporarily increase Python's recursion limit to handle cases where
    parsing a DAML-LF archive requires deeper recursion limits.
    """
    # noinspection PyPackageRequirements
    from google.protobuf.message import DecodeError
    from . import model as G

    current_time = time.time()

    prev_recursion_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(5000)
    archive_payload = G.ArchivePayload()
    try:
        archive_payload.ParseFromString(raw_bytes)
    except DecodeError:
        # noinspection PyPackageRequirements
        from google.protobuf.internal import api_implementation
        if api_implementation.Type() == 'cpp':
            LOG.error('Failed to decode metadata. This may be due to bugs in the native Protobuf')
            LOG.error('implementation as exposed through Python, so setting an environment')
            LOG.error('variable to force a non-native implementation may help work around this')
            LOG.error('problem:')
            LOG.error('    export PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION=python')
        raise
    finally:
        sys.setrecursionlimit(prev_recursion_limit)

    final_time = time.time()
    total_millis = (final_time - current_time) * 1000
    LOG.info('Parsed %s bytes of metadata in %2.f ms.', len(raw_bytes), total_millis)

    return archive_payload


def find_dependencies(
        metadatas_pb: 'Mapping[str, G.ArchivePayload]', existing_package_ids: 'Collection[str]') \
        -> 'Mapping[str, G.ArchivePayload]':
    """
    Return a topologically-sorted list of dependencies for the package IDs.

    :param metadatas_pb:
        The collection of gRPC ArchivePayload to re-order, indexed by package ID.
    :param existing_package_ids:
        Collection of package IDs that are to be treated as pre-existing.
    :return:
        A topologically-sorted dictionary of package IDs to ArchivePayload objects. Iterations
        through the dictionary are guaranteed to be in the correct order.
    """
    dependencies = defaultdict(set)
    for package_id, archive_payload in metadatas_pb.items():
        for module_pb in archive_payload.daml_lf_1.modules:
            for data_type_pb in module_pb.data_types:
                deps = None
                if data_type_pb.HasField('record'):
                    deps = find_dependencies_of_fwts(data_type_pb.record.fields)
                elif data_type_pb.HasField('variant'):
                    deps = find_dependencies_of_fwts(data_type_pb.variant.fields)
                if deps is not None:
                    deps.difference_update(existing_package_ids)
                    dependencies[package_id].update(deps)

    m_pb = {}
    sorted_package_ids = toposort_flatten(dependencies)
    for package_id in sorted_package_ids:
        m_pb[package_id] = metadatas_pb[package_id]
    return m_pb


def find_dependencies_of_fwts(fwts_pb) -> 'Set[str]':
    dependencies = set()
    for fwt in fwts_pb:
        dependencies.update(find_dependencies_of_fwt(fwt))
    return dependencies


def find_dependencies_of_fwt(fwt_pb) -> 'Collection[str]':
    return find_dependencies_of_type(fwt_pb.type)


def find_dependencies_of_type(type_pb) -> 'Collection[str]':
    t = type_pb.WhichOneof('Sum')  # type: str
    if t == 'prim':
        dependencies = set()  # type: Set[str]
        for arg in type_pb.prim.args:
            dependencies.update(find_dependencies_of_type(arg))
        return sorted(dependencies)
    elif t == 'con':
        dependencies = set()  # type: Set[str]
        if type_pb.con.tycon.module.package_ref.WhichOneof('Sum') == 'package_id':
            dependencies.add(type_pb.con.tycon.module.package_ref.package_id)
        for arg in type_pb.con.args:
            dependencies.update(find_dependencies_of_type(arg))
        return sorted(dependencies)
    elif t == 'var':
        dependencies = set()  # type: Set[str]
        for arg in type_pb.var.args:
            dependencies.update(find_dependencies_of_type(arg))
        return sorted(dependencies)
    elif t == 'fun':
        dependencies = set()  # type: Set[str]
        for arg in type_pb.fun.params:
            dependencies.update(find_dependencies_of_type(arg))
        dependencies.update(find_dependencies_of_type(type_pb.fun.result))
        return dependencies
    elif t == 'forall':
        return find_dependencies_of_type(type_pb.forall.body)
    elif t == 'tuple':
        return find_dependencies_of_fwts(type_pb.tuple.fields)
    else:
        LOG.warning('Unknown DAML-LF Type: %s (when evaluating %s)', t, type_pb)
        return ()


def parse_daml_metadata_pb(package_id: str, metadata_pb: Any) -> 'PackageStore':
    """
    Parse the contents of the given DAML-LF archive.

    :param package_id:
        The package to associate "self"-modules to.
    :param metadata_pb:
        The Protobuf-generated Archive object.
    :return:
        A :class:`PackageStore` with additional entries resulting from the parse of this archive.
    """
    from ...damlast.pb_parse import ProtobufParser

    parser = ProtobufParser(package_id)
    package = parser.parse_Package(metadata_pb.daml_lf_1)

    psb = PackageStoreBuilder()
    psb.add_archive(Archive(package_id, package))

    for module in package.modules:
        current_module = ModuleRef(package_id, module.name.segments)
        for vv in module.values:
            vt = ValueReference(current_module, vv.name_with_type.name)
            psb.add_value(vt, vv.expr)

        for dt in module.data_types:
            tt = create_data_type(current_module, dt)
            if isinstance(tt, (RecordType, VariantType)):
                psb.add_type(tt.name, tt)
            else:
                LOG.warning('Unexpected non-complex type')

        for template_pb in module.templates:
            tt = TypeReference(current_module, template_pb.tycon.segments)
            data_type = psb.get_type(tt)
            if isinstance(data_type, RecordType):
                psb.add_template(Template(
                    data_type=data_type,
                    choices=[
                        TemplateChoice(
                            c.name, c.consuming, get_old_type(c.arg_binder.type), c.controllers)
                        for c in template_pb.choices],
                    observers=template_pb.observers,
                    signatories=template_pb.signatories,
                    agreement=template_pb.agreement,
                    ensure=template_pb.precond))
            elif data_type is None:
                LOG.warning('The template %s did not have a corresponding data definition; '
                            'it will be ignored', tt)
            else:
                LOG.warning(
                    'The template %s was of type %s; only records are supported for templates',
                    tt, data_type)

    LOG.debug('Fully registered all types for package ID %r', package_id)
    return psb.build()


def create_data_type(current_module_ref: 'ModuleRef', dt: 'DefDataType') \
        -> 'Union[RecordType, VariantType, ScalarType]':
    from ...damlast.types import get_old_type

    type_vars = tuple(TypeVariable(type_var.var) for type_var in dt.params)
    tt = TypeReference(current_module_ref, tuple(dt.name.segments))

    if dt.record is not None:
        d = OrderedDict()
        for fwt in dt.record.fields:
            d[fwt.field] = get_old_type(fwt.type)
        return RecordType(NamedArgumentList(d.items()), tt, type_vars, TypeAdjective.USER_DEFINED)
    elif dt.variant is not None:
        d = OrderedDict()
        for fwt in dt.variant.fields:
            d[fwt.field] = get_old_type(fwt.type)
        return VariantType(NamedArgumentList(d.items()), tt, type_vars, TypeAdjective.USER_DEFINED)
    else:
        return SCALAR_TYPE_UNIT



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .plugins_base import Plugin
from .capture.plugin_capture import LedgerCapturePlugin



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Simple callbacks manager.
"""

import asyncio
import inspect
import reprlib
from asyncio import ensure_future, Future, get_event_loop
from functools import wraps
from typing import Dict, Optional, List, Callable, Any, Awaitable

from dazl.util.asyncio_util import propagate
from ..util.typing import safe_cast

from .. import LOG


class CallbackManager:
    """
    A manager for callbacks, either normal ones or coroutines.
    """
    def __init__(self, normalize_name_fn=None):
        self._records = dict()  # type: Dict[str, CallbackRecord]
        if normalize_name_fn is not None:
            self._normalize_name = normalize_name_fn

    def add_listener(self, name, callback: Callable[..., Any], filter_fn=None, optional=False) -> 'CallbackDisconnector':
        """
        Add a listener to be invoked whenever ``for_listeners(name)`` is called.

        :param name:
            The name of the listener to attach.
        :param callback:
            A normal function or a coroutine function that returns something that resembles a
            Command.
        :param filter_fn:
            An optional callback that returns `True` or `False` on whether the corresponding
            callback should be invoked. This cannot be a coroutine function.
        :param optional:
            Track whether this callback has actually been invoked or not.
        """
        if not callable(callback):
            raise ValueError('callback must be callable')

        name = self._normalize_name(name)
        record = self._records.get(name)
        if record is None:
            record = CallbackRecord()
            self._records[name] = record
        record.append(callback, filter_fn)
        return CallbackDisconnector(record, callback)

    def remove_listener(self, name, callback) -> bool:
        name = self._normalize_name(name)
        record = self._records.get(name)
        if record is not None:
            try:
                record.remove(callback)
                return True
            except ValueError:
                pass

        return False

    def for_listeners(self, name) -> Callable[..., List['CallbackRecord']]:
        """
        Returns a proxy callback that takes parameters that are passed to each callback as is.
        Any return values are gathered and returned as a list of Future in the same order as the
        original callbacks.
        """
        try:
            hash(name)
        except TypeError:
            LOG.error('Tried to use %s as a key for looking up event handlers.', name)
            raise

        name = self._normalize_name(name)
        cb = self._records.get(name)
        return cb if cb is not None else lambda *_, **__: []

    def rewrite_keys(self, rewrite_fn):
        self._records = {rewrite_fn(key): value for key, value in self._records.items()}

    def keys(self):
        return self._records

    def __repr__(self):
        return f'CallbackManager(keys={reprlib.repr(sorted(self.keys()))})'

    def __bool__(self):
        return bool(self._records)

    _normalize_name = str


class EventHandlerTracker:
    """
    Track the internal state of a callback that is subscribed as an event handler.

    The primary purpose of this is to ensure a callback is never called reentrantly, as it may
    introduce race conditions. The implementation assumes it is being used from ``asyncio`` so
    no locking is required.
    """
    __slots__ = ('original_callback', 'callback', 'pendings')

    def __init__(self, callback: Callable[..., Any], filter_fn=None):
        self.original_callback = callback
        self.callback = _normalize_callback_implementation(callback)
        self.pendings = []  # type: List[_StartableCallable]

    def invoke_eventually(self, *args, **kwargs) -> Awaitable[Any]:
        """
        Either invoke the underlying callback immediately and return a :class:`Future`, or place
        the invocation on an internal queue and invoke the underlying callback as soon as the
        inflight callback is fully invoked.

        :param args: Positional arguments to the callback.
        :param kwargs: Keyword arguments to the callback.
        :return: A Future that is resolved when the underlying callback is invoked.
        """
        sc = _StartableCallable(lambda: self.callback(*args, **kwargs), start=not self.pendings, name=repr(self.callback))
        self.pendings.append(sc)
        sc.add_done_callback(self._maybe_execute_next)
        return sc

    def _maybe_execute_next(self, fut: Future) -> None:
        # get the next callback to be invoked; if the list is empty, then we nothing to do
        try:
            sc = self.pendings.pop(0)
        except IndexError:
            LOG.warning('Received a notification that a Future finished that we don\'t care about')
            return

        if sc._future is not fut:
            LOG.error('Awaitables are being triggered in an incorrect order!')

        if self.pendings:
            self.pendings[0].start()


class _StartableCallable:
    """
    A :class:`Future`-like object that invokes a callback to start a computation at a later point
    in time.
    """
    _asyncio_future_blocking = False

    def __init__(self, cb: Callable[[], Future], start=False, name=None):
        if not callable(cb):
            raise ValueError('cb must be callable')

        if start:
            self._future = cb()
            self._callback = None
        else:
            self._future = get_event_loop().create_future()
            self._callback = cb

        # This is required to be compatible with the interface of a Future.
        self._loop = self._future._loop
        self._name = name or repr(cb)

    def start(self) -> None:
        """
        Actually invoke the underlying function (but only if it hasn't been invoked already).
        """
        cb = self._callback
        if cb is not None:
            self._callback = None
            propagate(from_=ensure_future(cb()), to=self._future)

    def started(self) -> bool:
        return self._callback is None

    def cancelled(self) -> bool:
        return self._future.cancelled()

    def cancel(self) -> bool:
        self._callback = None
        return self._future.cancel()

    def result(self) -> Any:
        return self._future.result()

    def done(self) -> bool:
        return self._future.done()

    def exception(self) -> Optional[BaseException]:
        return self._future.exception()

    def add_done_callback(self, callback: Callable[[Future], Any]) -> None:
        self._future.add_done_callback(callback)

    def __await__(self):
        return self._future.__await__()

    def __repr__(self):
        state = self._future._state if self.started() else 'NOT_STARTED'
        return f'StartableFuture({self._name!r}, state={state})'


def _normalize_callback_implementation(callback: Callable[..., Any]) -> Callable[..., Future]:
    """
    Convert the allowable callback signatures to a function that
    :return:
    """
    if not callable(callback):
        raise ValueError('callback must be callable')

    @wraps(callback)
    def invoke_sync(*args, **kwargs) -> Future:
        loop = asyncio.get_event_loop()

        try:
            result = callback(*args, **kwargs)
            if isinstance(result, asyncio.Future):
                return result
            else:
                future = loop.create_future()
                future.set_result(result)
                return future
        except Exception as ex:
            future = loop.create_future()
            future.set_exception(ex)
            return future

    if inspect.iscoroutinefunction(callback):
        return lambda *args, **kwargs: ensure_future(callback(*args, **kwargs))
    else:
        return invoke_sync


class CallbackRecord:
    """
    Manages the internal list of callbacks for a particular event.
    """

    __slots__ = ('trackers',)

    def __init__(self):
        self.trackers = []  # type: List[EventHandlerTracker]

    def __call__(self, *args, **kwargs) -> List[Awaitable[Any]]:
        return [tracker.invoke_eventually(*args, **kwargs) for tracker in self.trackers]

    def append(self, callback: Callable[..., Any], filter_fn=None) -> None:
        self.trackers.append(EventHandlerTracker(callback, filter_fn))

    def remove(self, callback) -> bool:
        """
        Remove an existing callback from the list of callbacks that will be invoked whenever
        this record is called.
        """
        for i, tracker in enumerate(self.trackers):
            if tracker.original_callback == callback:
                del self.trackers[i]
                return True
        return False


class InflightTracker:

    __slots__ = ('future', 'callback', 'args', 'kwargs')

    @classmethod
    def immediate(cls, callback, args, kwargs):
        tracker = InflightTracker(None, callback, args, kwargs)
        tracker.future = tracker.start()
        return tracker

    @classmethod
    def deferred(cls, callback, args, kwargs):
        future = asyncio.get_event_loop().create_future()
        return InflightTracker(future, callback, args, kwargs)

    def __init__(self, future: Optional[Future], callback, args, kwargs):
        self.future = safe_cast(Future, future) if future is not None else None
        if callable(callback):
            self.callback = callback
        else:
            raise ValueError('callable required here')
        self.args = args
        self.kwargs = kwargs

    def start(self):
        LOG.debug("We're starting a thing...")
        return asyncio.ensure_future(self.callback(*self.args, **self.kwargs))


class CallbackDisconnector:
    __slots__ = ('record', 'callback')

    def __init__(self, record, callback):
        self.record = record
        self.callback = callback

    """
    Returned from ``CallbackManager.add_listener`` as an easy way of disconnecting the specified
    event handler.
    """
    def disconnect(self) -> bool:
        return self.record.remove(self, self.callback)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from asyncio import gather
from argparse import ArgumentParser
from .. import LOG, Network, write_acs
from ._base import CliCommand
from ..plugins import LedgerCapturePlugin
from ..client.config import configure_parser, get_config


class ListAllCommand(CliCommand):
    name = 'ls'

    def parser(self) -> ArgumentParser:
        arg_parser = ArgumentParser('dazl ls')

        configure_parser(arg_parser, config_file_support=True)
        arg_parser.add_argument('--format', '--fmt', '-F', type=str, default=LedgerCapturePlugin.DEFAULT_FORMATTER_NAME)
        arg_parser.add_argument('--template-filter', '-T', type=str)
        arg_parser.add_argument('--all', '-A', action='store_true')
        return arg_parser

    def execute(self, args) -> int:
        fmt = args.format
        template_filter = [template.strip() for template in args.template_filter.split(',')] \
                          if args.template_filter is not None else None
        include_archived = bool(args.all)
        LOG.debug('Executing an ls...')

        final_config = get_config(args, config_file_support=True)

        network = Network()
        network.set_config(final_config)

        global_ready = gather(*[network.aio_party(party).ready() for party in args.parties])
        network.run_until_complete(self._main(network, global_ready, fmt, include_archived))
        return 0

    async def _main(self, network, global_ready, fmt, include_archived):
        import sys
        await global_ready

        write_acs(sys.stdout, network, fmt=fmt, include_archived=include_archived)
        network.shutdown()



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
:mod:`dazl.damlsdk` package
===========================

Module that exposes general utility functions around the DAML SDK Assistant.
"""

from ._errors import SDKComponentNotFoundError, SDKComponentUnknownError
from .fetch import ensure_sdk_component, sdk_component_path



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from aiohttp import web


def get_app() -> 'web.Application':
    from .metrics import build_routes as build_metrics_routes
    app = web.Application()
    app.add_routes(build_metrics_routes())
    return app



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import get_event_loop
from argparse import ArgumentParser

from .. import sandbox
from ..util.dar import DamlcPackageError
from ._base import CliCommand


class SandboxCommand(CliCommand):
    name = 'sandbox'
    hidden = True

    def parser(self):
        arg_parser = ArgumentParser('dazl sandbox')
        arg_parser.add_argument(
            'file', metavar='FILE', help='A .daml file to compile into a package, or a .dar file')
        arg_parser.add_argument(
            '--port', metavar='PORT', default=7600, type=int)
        arg_parser.add_argument(
            '--log-level')
        return arg_parser

    def execute(self, args):
        try:
            with sandbox(args.file, args.port):
                get_event_loop().run_forever()
        except DamlcPackageError as ex:
            return ex.exit_code



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains a plugin for capturing data off the ledger and outputting it.
"""

import sys

from functools import partial
from typing import Collection, Iterable, TextIO, Optional, Union, TYPE_CHECKING

from .fmt_base import DEFAULT_FORMATTER_NAME, get_formatter
from .model_capture import LedgerCapture
from ..plugins_base import Plugin
from ...model.core import Party

if TYPE_CHECKING:
    from ...client import LedgerClientManager, Network


class LedgerCapturePlugin(Plugin):
    """
    Plugin that passively listens to the event stream on all parties. Call :meth:`dump_all` to
    render what it sees.
    """

    DEFAULT_FORMATTER_NAME = DEFAULT_FORMATTER_NAME

    @classmethod
    def stdout(cls, **kwargs) -> 'LedgerCapturePlugin':
        """
        Return a :class:`LedgerCapturePlugin` that writes its output to stdout.
        """
        return cls(sys.stdout, False, **kwargs)

    @classmethod
    def stderr(cls, **kwargs) -> 'LedgerCapturePlugin':
        """
        Return a :class:`LedgerCapturePlugin` that writes its output to stderr.
        """
        return cls(sys.stderr, False, **kwargs)

    @classmethod
    def to_file(cls, path, **kwargs) -> 'LedgerCapturePlugin':
        """
        Return a :class:`LedgerCapturePlugin` that writes its output to a file.

        :param path: The file to write to.
        """
        return cls(open(path, 'w'), True, **kwargs)

    def __init__(self,
                 buf: TextIO,
                 buf_close: bool,
                 template_filter: Optional[Iterable[str]]=None,
                 include_archived: bool=False):
        """
        Initialize a :class:`LedgerCapturePlugin`.

        :param buf:
            The buffer to write ledger state to.
        :param buf_close:
            ``True`` to close ``buf`` after fully dumping the contents of the ledger;
            ``False`` to keep ``buf`` open.
        :param template_filter:
            A list of ``str`` of template names. The default value is ``None``, which includes all
            templates.
        :param include_archived:
            ``True`` to include archived contracts in the output;
            otherwise, ``False`` (the default).
        """
        self.buf = buf
        self.buf_close = buf_close
        self.template_filter = template_filter
        self.include_archived = include_archived
        self.manager = None
        self.store = None

    def install(self, manager: 'LedgerClientManager'):
        """
        Monitor the events off of a ledger client.

        :param manager:
            The :class:`LedgerClientManager` to install event handlers into.
        """
        self.manager = manager
        self.manager.on_init_metadata(partial(setattr, self, 'store'))

    def dump_all(self, fmt=None, parties=None, include_archived=None, **kwargs) -> None:
        """
        Write all entries captured by the plugin to the provided buffer.

        :param fmt:
            Format of the output; currently either ``"json"`` or ``"pretty"``.
        :param parties:
            An optional list of parties. If unspecified, the list is taken from the list of parties
            connected to the client.
        :param include_archived:
            ``None`` to use the default value passed in the constructor; ``True`` to include
            archived contracts in the output; otherwise, ``False``.
        :param kwargs:
            Parameters passed directly to the formatter.
        """
        formatter = get_formatter(fmt)

        if self.manager is None:
            lines = formatter.format_error('Plugin not initialized')
        else:
            if parties is None:
                parties = sorted(self.manager.parties())

            include_archived = include_archived if include_archived is not None \
                else self.include_archived

            capture = LedgerCapture()
            capture.store = self.store
            for party in self.manager.parties():
                client = self.manager.client(party)

                if include_archived:
                    for cid, (cdata, state) in client.select('*', include_archived=True).items():
                        if cdata is not None:
                            capture.capture(party, cid, cdata)
                        if not state:
                            capture.capture(party, cid, None)
                else:
                    for cid, cdata in client.select('*').items():
                        capture.capture(party, cid, cdata)

            lines = formatter.format_entries(
                capture=capture,
                parties=parties,
                entries=capture,
                **kwargs)

        for line in lines:
            self.buf.write(line + '\n')


def write_acs(
        buf: 'TextIO',
        network: 'Network',
        fmt: 'Optional[str]' = None,
        parties: 'Collection[Union[str, Party]]' = None,
        include_archived: bool = False,
        **kwargs) \
        -> None:
    """
    Write all entries captured by the plugin to the provided buffer.

    :param buf:
        The buffer to write to.
    :param network:
        The network to write data from.
    :param fmt:
        Format of the output; currently either ``"json"`` or ``"pretty"``.
    :param parties:
        An optional list of parties. If unspecified, the list is taken from the list of parties
        connected to the client.
    :param include_archived:
        ``True`` to include archived contracts in the output; otherwise, ``False``.
    :param kwargs:
        Parameters passed directly to the formatter.
    """
    formatter = get_formatter(fmt)

    if parties is None:
        parties = sorted(network.parties())

    capture = LedgerCapture()
    capture.store = network.simple_global().metadata().store
    for party in parties:
        # TODO: This is a hack because ACS calls on a simple_party instance don't really work once
        #   the loop is closed.
        client = network.aio_party(party)

        for cxdata in client.find_historical('*'):
            if cxdata.active:
                capture.capture(party, cxdata.cid, cxdata.cdata, cxdata.effective_at)
            elif include_archived:
                capture.capture(party, cxdata.cid, None, None)

    lines = formatter.format_entries(
        capture=capture,
        parties=parties,
        entries=capture,
        **kwargs)

    for line in lines:
        buf.write(line + '\n')



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains classes and methods for managing a collection of clients.
"""

from asyncio import get_event_loop
from datetime import datetime
from functools import wraps
from typing import Collection, Union

from dataclasses import asdict

from .client_participant import ParticipantLedgerClient
from .config import get_config, parse_kwargs, validate_config, NetworkConfig
from ._base_model import ExitCode, LedgerRun
from ._run_level import RunState
from ._network_client_impl import _NetworkImpl
from ..model.core import RunLevel, Party
from ..model.reading import InitEvent
from ..plugins.plugins_base import Plugin
from ..util.prim_types import PrimitiveTypeConverter


def create_client(*config, **kwargs):
    """
    Create a :class:`LedgerClientManager` for the configuration settings.

    :param config:
        Instances of :class:`LedgerConfiguration`` objects to merge together.
    :param kwargs:
        Configuration options that are accepted either at the global level or at the party level.
    :return:
        An instance of :class:`LedgerClientManager`.
    """
    cfg = parse_kwargs(*config, **kwargs)
    return LedgerClientManager(cfg)


class LedgerClientManager:
    """
    Keep-alive manager for open client instances. This is the primary entry point for interacting
    with the Ledger.
    """

    @classmethod
    def for_args(cls, args):
        return cls(get_config(args))

    def __init__(self, config: 'NetworkConfig'):
        # import warnings
        # warnings.warn('LedgerClientManager will be removed in version 6.0.0. Please switch to '
        #               'LedgerNetwork for managing instances of clients for parties that are '
        #               'connected to the ledger.', DeprecationWarning)
        self._config = validate_config(config)
        self._primitive_type_converter = PrimitiveTypeConverter()
        self._impl = _NetworkImpl()

        for party_config in config.parties:
            self.new_client(**asdict(party_config))

    def __enter__(self):
        """
        Attempt to initialize the committer node. Otherwise do nothing.
        """
        return self

    def __exit__(self, typ, value, traceback):
        """
        Attempt to stop the client manager and all connected clients.
        """

    def register(self, plugin, keep_alive=True):
        """
        Register a plugin with the client.

        If the plugin implements Plugin, then its methods are called as documented. Otherwise, the
        plugin is immediately called with this :class:`LedgerClientManager` as its only argument.
        It should take this opportunity to register event handlers as required to do its work.

        Additionally, if an ``asyncio.Future`` or a coroutine is returned, it is expected to
        indicate the lifetime of the extension:

          - If the future is resolved successfully, the extension is expected to no longer
            interact with the ledger, nor does it require any more "time" to run. If ``keep_alive``
            is ``True``, this pending ``Future`` will additionally prevent the ledger client from
            completing.

          - If the future is resolved with a failure, the entire application is considered to be in
            an error state, and the client is shut down as quickly as is reasonably possible.
        """
        if plugin is not None:
            if isinstance(plugin, Plugin):
                plugin.install(self)
            else:
                raise ValueError('must be a Plugin')

    def stop(self) -> None:
        """
        Wait for all managed clients to finish any outstanding work, and then stops them.
        """
        self._impl.shutdown()

    def terminate(self) -> None:
        """
        Fairly rudely kill the application.
        """
        import sys
        self._impl.abort()
        sys.exit(3)

    def client(self, party_name: Union[str, Party]) -> ParticipantLedgerClient:
        """
        Return an existing client for a party. If none does not exist, :class:`KeyError` is thrown.
        """
        client = self._impl.party_impl(party_name, ParticipantLedgerClient)
        if client is None:
            raise KeyError(party_name)
        return client

    def new_client(self, party: Union[str, Party], **config) -> 'ParticipantLedgerClient':
        """
        Create a new client for the specified party, or if one already exists, augment the provided
        configuration parameters to the existing client.

        :param party: The party.
        :param config: Key-value pairs that represent configuration parameters.
        :return: An instance of :class:`LedgerClientManager`.
        """
        client = self._impl.party_impl(party, ParticipantLedgerClient)
        client.set_config(**config)
        return client

    def _handle_client_init_metadata(self, metadata):
        # TODO: This needs to be thought through more carefully; each cilent fetches its own
        # metadata and this just gets run when the "first" one gets its metadata. Should we wait
        # until all readers have heard their metadata, and then compare? We already block pending
        # writes until all clients are signaled...
        local_init_metadata_callbacks = self._init_metadata_callbacks[:]
        self._init_metadata_callbacks = []
        for init_metadata_callback in local_init_metadata_callbacks:
            init_metadata_callback(metadata)

    def _handle_client_error(self, error):
        for err_callback in self._error_callbacks:
            err_callback(error)

    async def get_time(self) -> datetime:
        return await self._impl.get_time()

    async def set_time(self, new_datetime):
        return await self._impl.set_time(new_datetime)

    def on_init(self, callback):
        # TODO: Probably need to rewrite this event
        self._impl.add_event_handler('init', callback, self)

    def on_init_metadata(self, callback):
        @wraps(callback)
        def rewrite(event: InitEvent) -> None:
            callback(event.package_store)
        self._impl.add_event_handler('init', rewrite, self)

    def on_done(self, callback):
        # TODO: Add deprecation warning
        pass

    def on_error(self, callback):
        # TODO: Add deprecation warning
        pass

    def run_until_complete(self, *futures) -> LedgerRun:
        """
        Start all clients, and terminate the event loop once work is done. This is primarily meant
        to be used when writing a script that is intended to do a specific thing and terminate (such
        as initialization scripts, run a test, or perform some specific action), as opposed to a
        long-lived application that continually acts on contracts that appear on the ledger.

        Specifically, the following is done:

        * First, submit all commands that have been queued up, either by prior calls to
          :meth:`ParticipantLedgerClient.submit` or as commands returned via an
          :meth:`ParticipantLedgerClient.on_ready`
        * Wait for those commands to have been fully acted on by the ledger (either success or
          failure). In the case of success, commands will typically cause contract created or
          contract archived events.
        * If there are callbacks registered to handle those events, they can, in turn, also produce
          commands to be sent to the ledger as well. If so, additionally wait for those commands to
          be completed as well.
        * Once there are no more commands awaiting completion, stop the client.

        If you need to control the event loop yourself, call the :meth:`main` coroutine method
        instead.
        """
        loop = get_event_loop()
        return loop.run_until_complete(self.main(False, *futures))

    def run_forever(self) -> LedgerRun:
        """
        Start all clients. Run until the process is sent a SIGINT signal or CTRL-C is pressed on
        the keyboard. This is primarily meant to be used when writing a long-lived application that
        continually acts on contracts that appear on the ledger, as opposed to a script that is
        intended to do a specific thing and terminate (such as initialization scripts, run a test,
        or perform some specific action).

        If you need to control the event loop yourself, call the :meth:`main` coroutine instead.
        """
        loop = get_event_loop()
        return loop.run_until_complete(self.main(True))

    def parties(self) -> Collection[Party]:
        """
        Return a snapshot of the set of parties that exist right now.
        """
        return self._impl.parties()

    ################################################################################################
    # CORE COROUTINES

    async def main(self, keep_alive: bool = True, *futures) -> LedgerRun:
        """
        The main coroutine that originates all client activity. This method can be used as an
        alternative to starting and stopping the client through ``run_until_complete`` and/or
        ``run_forever``.

        :param keep_alive:
            If ``True``, keep the coroutine running until :meth:`stop` is explicitly called;
            otherwise ``False`` to end the coroutine when all commands and their follow-ups have
            been processed.
        """
        run_state = RunState(RunLevel.RUN_FOREVER if keep_alive else RunLevel.RUN_UNTIL_IDLE)
        await self._impl.aio_run(*futures, run_state=run_state)
        return LedgerRun(
            exit_code=ExitCode.SUCCESS,
            block_start_height=None,
            block_end_height=None)




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains classes and methods for interacting with a participant node.
"""
import warnings
from datetime import datetime
from functools import wraps
from typing import Any, Awaitable, Callable, Optional, Tuple, Union

from ._party_client_impl import _PartyClientImpl
from ..model.core import ContractId, ContractData, ContractMatch, ContractsHistoricalState, \
    ContractsState, Party
from ..model.reading import EventKey, ContractCreateEvent
from ..model.types_store import PackageStore
from ..util.asyncio_util import await_then


class ParticipantLedgerClient:
    """
    This class is deprecated and will be removed in version 6.0.0. Prefer :class:`AIOPartyClient`
    if you want to use ``asyncio`` or :class:`SimplePartyClient` if you want to use a blocking model
    instead.
    """
    def __init__(self, party_impl: _PartyClientImpl):
        warnings.warn('ParticipantLedgerClient will be removed in version 6.0.0. Please switch to '
                      'AIOPartyClient (if you wish to continue to use asyncio) or '
                      'SimplePartyClient (if you wish to use a blocking model instead),',
                      DeprecationWarning)
        self._impl = party_impl

    @property
    def party(self) -> Party:
        return self._impl.party

    @property
    def party_name(self) -> Party:
        warnings.warn('`party_name` is a deprecated property name. Please us `party` instead.`',
                      DeprecationWarning)
        return self._impl.party

    def set_config(self, **kwargs):
        return self._impl.set_config(**kwargs)

    # region Read Methods (ACS)

    def find_one(self, template, match):
        return self.select_first(template, match)

    def find(self, *args, **kwargs):
        return self._impl.find(*args, **kwargs)

    def find_active(self, *args, **kwargs):
        return self._impl.find_active(*args, **kwargs)

    def find_historical(self, *args, **kwargs):
        return self._impl.find_historical(*args, **kwargs)

    def find_nonempty(self, *args, **kwargs):
        return self._impl.find_nonempty(*args, **kwargs)

    def select(self, template_name: str, match: ContractMatch = None, include_archived=False) \
            -> Union[ContractsState, ContractsHistoricalState]:
        """
        Immediately return data from the current active contract set.

        The contents of this ACS are guaranteed to be present (or removed) in the current
        transaction _before_ processing any corresponding ``on_created`` or ``on_archived``
        callbacks for this party. The ACS is populated _before_ processing any ``on_ready``
        callbacks.

        This method raises an error if ACS tracking has been disabled on this client.

        :param template_name:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :param include_archived:
            ``True`` to include values that have been archived. When this is set, the returned
            dictionary's values is a tuple of contract data and the active state (either ``True`` or
            ``False``).
        :return:
            A ``dict`` whose keys are :class:`ContractId` and values are corresponding contract
            data that match the current query.
        """
        if include_archived:
            return self._impl.find_historical(template_name, match)
        else:
            return self._impl.find_active(template_name, match)

    def select_first(self, template_name: str, match: ContractMatch = None) \
            -> Awaitable[Tuple[ContractId, ContractData]]:
        """
        Convenience method that returns the first occurrence of a matching contract.
        """
        def extract_state(contracts):
            return next(iter(contracts.items()))
        return await_then(self._impl.find_nonempty(template_name, match), extract_state)

    def select_nonempty(self, template_name: str, match: ContractMatch = None, min_count: int = 1) \
            -> Awaitable[ContractsState]:
        """
        Return data from the current active contract set when at least some amount of rows exist in
        the active contract set.

        :param template_name:
            The name of the template to fetch data from.
        :param match:
            An optional dictionary whose keys are matched against corresponding field values.
        :param min_count:
            The minimum number of rows to return. The default value is 1.
        :return:
            A ``Future`` that is resolved with a ``dict`` whose keys are :class:`ContractId` and
            values are corresponding contract data that match the current query.
        """
        return self._impl.find_nonempty(template_name, match, min_count=min_count)

    # endregion

    # region Write Methods

    def submit(self, commands, ignore_errors=False, workflow_id=None):
        """
        Submit the set of commands for processing.

        :param commands:
            A single command or a list of commands.
        :param ignore_errors:
            ``True`` if failure to submit a command should not prevent termination of the program.
        :return:
            A future that resolves when the commands have made it to the ledger _or_ an error
            occurred when trying to process them.
        """
        return self._impl.write_commands(commands, ignore_errors, workflow_id=workflow_id)

    def submit_create(self, template_name: str, arguments: Optional[dict] = None, workflow_id=None):
        """
        Submit a single create command. Equivalent to calling :meth:`submit` with a single
        ``create``.

        :param template_name:
            The name of the template.
        :param arguments:
            The arguments to the create (as a ``dict``).
        :return:
            A future that resolves when the command has made it to the ledger _or_ an error
            occurred when trying to process them.
        """
        from .. import create
        return self.submit(create(template_name, arguments), workflow_id=workflow_id)

    def submit_exercise(self, cid: ContractId, choice_name: str, arguments: Optional[dict] = None, workflow_id=None):
        """
        Submit a single exercise choice. Equivalent to calling :meth:`submit` with a single
        ``exercise``.

        :param cid:
            The :class:`ContractId` on which a choice is being exercised.
        :param choice_name:
            The name of the choice to exercise.
        :param arguments:
            The arguments to the create (as a ``dict``). Can be omitted (``None``) for no-argument
            choices.
        :return:
            A future that resolves when the command has made it to the ledger _or_ an error
            occurred when trying to process them.
        """
        from .. import exercise
        return self.submit(exercise(cid, choice_name, arguments), workflow_id=workflow_id)

    # endregion

    # region Other Methods

    async def get_time(self) -> datetime:
        """
        Return the current time on the remote server. Also advances the local notion of time if
        required.
        """
        return await self._impl.get_time()

    def set_time(self, new_datetime: datetime) -> Awaitable[None]:
        """
        Set the current time on the ledger. This is only supported if the ledger supports time
        manipulation.
        """
        return self._impl.set_time(new_datetime)

    # endregion

    # region Event Handlers

    def ready(self) -> Awaitable[None]:
        """
        An Awaitable signaled once the ledger client has caught up to the current head and is
        ready to send commands.
        """
        return self._impl.ready()

    # noinspection PyShadowingBuiltins
    def add_event_handler(self, event_key, event_handler, filter=None) -> None:
        """
        Register an event handler.
        """
        self._impl.add_event_handler(event_key, event_handler, filter, self)

    def on_init(self, callback: 'Callable[[], Any]') -> None:
        """
        Called once the ledger client has been instructed to begin, but before any network activity
        is started.
        """
        # Historically on_init callbacks took no parameters
        for key in EventKey.init():
            self.add_event_handler(key, lambda _: callback())

    def on_init_metadata(self, callback: 'Callable[[PackageStore], Any]') -> None:
        """
        Called once metadata has been discovered.
        """
        # Historically on_init_metadata callbacks took a single parameter that is the metadata of
        # the ledger, and additionally was treated as a separate event from init.
        for key in EventKey.init():
            self.add_event_handler(key, lambda evt: callback(evt.package_store))

    def on_ready(self, callback: 'Callable[[str, ParticipantLedgerClient], Any]') -> None:
        """
        Called once the ledger client has caught up to the current head and is ready to send
        commands, or called immediately if the client is already ready.

        :param callback:
            Callback that takes two arguments (party name and the instance of this client). If this
            callback returns follow-up commands, they are submitted.
        """
        for key in EventKey.ready():
            self.add_event_handler(key, lambda evt: callback(evt.party, evt.client))

    def on_event(self, event_kind, callback, optional=False) -> None:
        """
        No longer supported.
        """
        raise RuntimeError(
            'generic event listeners are no longer supported. '
            'Please register for specific events instead')

    # noinspection PyUnusedLocal
    def on_created(self, template, match=None, callback=None, optional=False) -> None:
        """
        Called whenever a contract of the specified template is created.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        :param callback:
            The callback to invoke whenever a matching template is created.
        :param optional:
            This parameter is no longer used.
        """
        if match is not None and callback is None:
            callback = match
            match = None

        @wraps(callback)
        def implementation(event: ContractCreateEvent):
            return callback(event.cid, event.cdata)

        if match is not None:
            @wraps(match)
            def match_impl(event: ContractCreateEvent):
                return match(event.cid, event.cdata)
        else:
            match_impl = None

        for key in EventKey.contract_created(True, template):
            self.add_event_handler(key, implementation, filter=match_impl)

    # noinspection PyUnusedLocal
    def on_archived(self, template, callback, optional=False) -> None:
        """
        Called whenever a contract of the specified template is archived.
        """
        for key in EventKey.contract_archived(True, template):
            self.add_event_handler(key, lambda evt: callback(evt.cid, evt.cdata))

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def on_done(self, callback):
        """
        Add a listener that gets called when this :class:`ParticipantLedgerClient` is terminated.
        """
        warnings.warn('The done event is no longer raised and does nothing.', DeprecationWarning)

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def on_error(self, callback):
        """
        Called on any error from the ledger.
        """
        warnings.warn('The error event is no longer raised and does nothing.', DeprecationWarning)

    # endregion

    def __repr__(self):
        return f'<ParticipantLedgerClient(party_name={self.party}>)'



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from datetime import datetime
from typing import Iterator, Optional

# noinspection PyPackageRequirements
from grpc import RpcError, StatusCode

from ... import LOG
from .pb_parse_event import to_datetime


def maybe_grpc_time_stream(time_service: 'G.TimeServiceStub', ledger_id: str) \
        -> Optional[Iterator[datetime]]:
    """
    Return an iterator over times that monotonically increases, or ``None`` if the TimeService is
    not implemented by the remote server.

    :param time_service:
    :param ledger_id:
    :return:
    """
    from . import model as G
    request = G.GetTimeRequest(ledger_id=ledger_id)

    time_stream = time_service.GetTime(request)
    try:
        return _TimeStream(time_stream)
    except StaticTimeUnsupportedError:
        LOG.debug('Operating in real-time mode.')
        return None


class _TimeStream(Iterator[datetime]):

    def __init__(self, time_stream):
        try:
            self.time_stream_iter = iter(time_stream)
            self.time_queue = [to_datetime(next(self.time_stream_iter).current_time)]
        except RpcError as ex:
            status_code = ex.code()
            if status_code == StatusCode.UNIMPLEMENTED:
                raise StaticTimeUnsupportedError()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.time_queue.pop()
        except IndexError:
            pass

        try:
            dt = to_datetime(next(self.time_stream_iter).current_time)
            LOG.debug('Got a new time: %s', dt)
            return dt
        except RpcError as ex:
            status_code = ex.code()
            if status_code == StatusCode.CANCELLED:
                # a cancelled code is fine; that just means the underlying time stream connection is
                # closed (which also means this function is done with its work)
                raise StopIteration
            else:
                raise


class StaticTimeUnsupportedError(Exception):
    pass



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains definitions for built-in DAML functions.
"""

import operator

from datetime import timedelta
from decimal import Decimal, ROUND_HALF_EVEN
from functools import reduce
from typing import Any, Callable, Optional, Union


_ONE = Decimal('1')
_ONE_THOUSAND = Decimal('1000')
_BUILTINS = dict()


class FunctionMetadata:
    name: str
    impl: Callable[..., Any]
    infix: bool

    def __init__(self, name, impl, infix):
        self.name = name
        self.impl = impl
        self.infix = infix


def define_builtin(symbol: str, fn: Callable[..., Any], infix: bool=False):
    func_metadata = FunctionMetadata(symbol, fn, infix)
    _BUILTINS[symbol] = func_metadata
    return func_metadata


def lookup_builtin(symbol: str) -> Optional['FunctionMetadata']:
    return _BUILTINS.get(symbol)


def foldl(func, acc, xs):
    return reduce(func, xs, acc)


def foldr(func, acc, xs):
    return reduce(lambda x, y: func(y, x), xs[::-1], acc)


def lookup(label, record):
    return record[label]


def extend(label: str, value: Any, record: dict) -> dict:
    new_record = dict(record)
    new_record[label] = value
    return new_record


def tag(label: str, value: Any) -> dict:
    return {label: value}


def round_(value):
    return Decimal(value).quantize(_ONE, ROUND_HALF_EVEN)


def exercises(parameter_count: int, choice_name: str, party: str, contract_id: 'ContractId', *arguments) -> 'Update':
    pass


def commit(party: str, update: 'Update') -> 'Scenario':
    pass


def n_arity(fn):
    return lambda *xs: reduce(fn, xs)


def from_relative_time(value: timedelta) -> Decimal:
    return Decimal(value._to_microseconds()) * _ONE_THOUSAND


def to_relative_time(value: Union[int, float]) -> timedelta:
    # TODO: This is a very lossy conversion; not clear what the best way to handle this is absent
    # a subclassing of ``timedelta``
    return timedelta(microseconds=float(Decimal(value) / _ONE_THOUSAND))


BUILTIN_ADD = define_builtin('+', n_arity(operator.add), infix=True)
BUILTIN_SUB = define_builtin('-', operator.sub, infix=True)
BUILTIN_MUL = define_builtin('*', n_arity(operator.mul), infix=True)
BUILTIN_DIV = define_builtin('/', operator.truediv, infix=True)
BUILTIN_POW = define_builtin('^', pow, infix=True)
BUILTIN_LT = define_builtin('<', operator.lt, infix=True)
BUILTIN_LE = define_builtin('<=', operator.le, infix=True)
BUILTIN_GT = define_builtin('>', operator.gt, infix=True)
BUILTIN_GE = define_builtin('>=', operator.ge, infix=True)
BUILTIN_EQ = define_builtin('==', operator.eq, infix=True)
BUILTIN_NE = define_builtin('/=', operator.ne, infix=True)
BUILTIN_AND = define_builtin('&&', n_arity(operator.and_), infix=True)
BUILTIN_OR = define_builtin('||', n_arity(operator.or_), infix=True)
BUILTIN_CONCAT = define_builtin('<>', n_arity(operator.add), infix=True)
# overlay

define_builtin('toText', str)
define_builtin('singleton', str)
define_builtin('unpack', list)
define_builtin('fromRelTime', from_relative_time)
define_builtin('toRelTime', to_relative_time)
define_builtin('fromInteger', Decimal)
define_builtin('toInteger', int)

define_builtin('round', round_)
define_builtin('divD', operator.truediv)
define_builtin('remD', operator.mod)

BUILTIN_NOT = define_builtin('not', operator.not_)
define_builtin('or', operator.or_)
define_builtin('and', operator.and_)

define_builtin('nil', lambda: [])
define_builtin('cons', lambda x, xs: xs + [x])
define_builtin('foldl', foldl)
define_builtin('foldr', foldr)

define_builtin('lookup', lookup)
define_builtin('extend', extend)
define_builtin('tag', tag)
define_builtin('exercises', exercises)
define_builtin('commit', commit)

define_builtin('create', lambda *args, **kwargs: None)
define_builtin('delete', lambda *args, **kwargs: None)
define_builtin('pure', lambda *args, **kwargs: None)
define_builtin('assert', lambda *args, **kwargs: None)
